@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,10 +16,9 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: commonfops.c,v 1.38 2005/10/20 16:24:41 cwright Exp $
+ *  $Id: commonfops.c,v 1.59 2006/06/02 03:29:26 dquigley Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 
 /* We only need this function here, but it could get promoted to unionfs.h, if
@@ -28,61 +28,55 @@ static inline void branchput_gen(int generation, struct super_block *sb,
 {
 	struct putmap *putmap;
 
-	unionfs_read_lock(sb);
-
 	if (generation == atomic_read(&stopd(sb)->usi_generation)) {
 		branchput(sb, index);
-		goto out;
+		return;
 	}
 
-	ASSERT(stopd(sb)->usi_firstputmap <= generation);
-	ASSERT(stopd(sb)->usi_lastputmap >= generation);
+	BUG_ON(stopd(sb)->usi_firstputmap > generation);
+	BUG_ON(stopd(sb)->usi_lastputmap < generation);
 
 	putmap =
 	    stopd(sb)->usi_putmaps[generation - stopd(sb)->usi_firstputmap];
-	PASSERT(putmap);
-	ASSERT(index >= 0);
-	ASSERT(index <= putmap->bend);
-	ASSERT(putmap->map[index] >= 0);
+	BUG_ON(index < 0);
+	BUG_ON(index > putmap->bend);
+	BUG_ON(putmap->map[index] < 0);
 	branchput(sb, putmap->map[index]);
 	if (atomic_dec_and_test(&putmap->count)) {
 		stopd(sb)->usi_putmaps[generation - stopd(sb)->usi_firstputmap]
 		    = NULL;
-		fist_dprint(8, "Freeing putmap %d.\n", generation);
+		dprint(PRINT_DEBUG, "Freeing putmap %d.\n", generation);
 		KFREE(putmap);
 	}
-
-      out:
-	unionfs_read_unlock(sb);
 }
 
-char *get_random_name(int size, unsigned char *name)
+static char *get_random_name(int size, unsigned char *name)
 {
 	int i;
 	int j;
 	unsigned char *tmpbuf = NULL;
 
-	if (size <= 4)
+	if (size <= WHLEN)
 		return NULL;
 
 	if (!name)
-		name = KMALLOC(size + 1, GFP_UNIONFS);
+		name = KMALLOC(size + 1, GFP_KERNEL);
 	if (!name) {
 		name = ERR_PTR(-ENOMEM);
 		goto out;
 	}
-	strncpy(name, ".wh.", 4);
+	strncpy(name, WHPFX, WHLEN);
 
-	tmpbuf = KMALLOC(size, GFP_UNIONFS);
+	tmpbuf = KMALLOC(size, GFP_KERNEL);
 	if (!tmpbuf) {
 		KFREE(name);
-		tmpbuf = ERR_PTR(-ENOMEM);
+		name = ERR_PTR(-ENOMEM);
 		goto out;
 	}
 
 	get_random_bytes((void *)tmpbuf, (size - 3) / 2);
 
-	j = 4;
+	j = WHLEN;
 	i = 0;
 	while ((i < (size - 3) / 2) && (j < size)) {
 		/* get characters in the 0-9, A-F range */
@@ -109,8 +103,8 @@ char *get_random_name(int size, unsigned char *name)
 
 }
 
-int copyup_deleted_file(struct file *file, struct dentry *dentry, int bstart,
-			int bindex)
+static int copyup_deleted_file(struct file *file, struct dentry *dentry,
+			       int bstart, int bindex)
 {
 	int attempts = 0;
 	int err;
@@ -130,7 +124,7 @@ int copyup_deleted_file(struct file *file, struct dentry *dentry, int bstart,
 		err = -ENOMEM;
 		if (!name)
 			goto out;
-
+		//XXX: Why do we do this every time? bstart never changes?
 		hidden_dentry = dtohd_index(dentry, bstart);
 
 		tmp_dentry = LOOKUP_ONE_LEN(name, hidden_dentry->d_parent,
@@ -171,31 +165,28 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 	int bindex, bstart, bend;
 	struct file *hidden_file;
 	struct dentry *hidden_dentry;
+	int size;
 
 	int err = 0;
 
 	print_entry(" file = %p", file);
 
-	PASSERT(file);
-	PASSERT(ftopd(file));
-
 	dentry = file->f_dentry;
 	lock_dentry(dentry);
-	PASSERT(dentry);
-	if (!unionfs_d_revalidate(dentry, 0) && !d_deleted(dentry)) {
+	sb = dentry->d_sb;
+	unionfs_read_lock(sb);
+	if (!unionfs_d_revalidate(dentry, NULL) && !d_deleted(dentry)) {
 		err = -ESTALE;
 		goto out;
 	}
-	fist_print_dentry("file revalidate in", dentry);
+	print_dentry("file revalidate in", dentry);
 
-	sb = dentry->d_sb;
-	PASSERT(sb);
 	sbgen = atomic_read(&stopd(sb)->usi_generation);
 	dgen = atomic_read(&dtopd(dentry)->udi_generation);
 	fgen = atomic_read(&ftopd(file)->ufi_generation);
-	if (sbgen > dgen) {
-		FISTBUG("The dentry is not up to date!\n");
-	}
+
+	BUG_ON(sbgen > dgen);
+
 	/* There are two cases we are interested in.  The first is if the
 	 * generation is lower than the super-block.  The second is if someone
 	 * has copied up this file from underneath us, we also need to refresh
@@ -211,6 +202,7 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 				fput(ftohf_index(file, bindex));
 			}
 		}
+
 		if (ftohf_ptr(file)) {
 			KFREE(ftohf_ptr(file));
 			ftohf_ptr(file) = NULL;
@@ -220,19 +212,12 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 		bstart = fbstart(file) = dbstart(dentry);
 		bend = fbend(file) = dbend(dentry);
 
-		if (sbmax(sb) > UNIONFS_INLINE_OBJECTS) {
-			int size =
-			    sizeof(struct file *) * (sbmax(sb) -
-						     UNIONFS_INLINE_OBJECTS);
-			ftohf_ptr(file) = KMALLOC(size, GFP_UNIONFS);
-			if (!ftohf_ptr(file)) {
-				err = -ENOMEM;
-				goto out;
-			}
-			memset(ftohf_ptr(file), 0, size);
+		size = sizeof(struct file *) * sbmax(sb);
+		ftohf_ptr(file) = KZALLOC(size, GFP_KERNEL);
+		if (!ftohf_ptr(file)) {
+			err = -ENOMEM;
+			goto out;
 		}
-		memset(ftohf_inline(file), 0,
-		       sizeof(struct file *) * UNIONFS_INLINE_OBJECTS);
 
 		if (S_ISDIR(dentry->d_inode->i_mode)) {
 			/* We need to open all the files. */
@@ -260,8 +245,6 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 		} else {
 			/* We only open the highest priority branch. */
 			hidden_dentry = dtohd(dentry);
-			PASSERT(hidden_dentry);
-			PASSERT(hidden_dentry->d_inode);
 			if (willwrite && IS_WRITE_FLAG(file->f_flags)
 			    && is_robranch(dentry)) {
 				for (bindex = bstart - 1; bindex >= 0; bindex--) {
@@ -314,16 +297,12 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 	/* Copyup on the first write to a file on a readonly branch. */
 	if (willwrite && IS_WRITE_FLAG(file->f_flags)
 	    && !IS_WRITE_FLAG(ftohf(file)->f_flags) && is_robranch(dentry)) {
-		fist_dprint(3,
-			    "Doing delayed copyup of a read-write file on a read-only branch.\n");
+		dprint(PRINT_DEBUG,
+		       "Doing delayed copyup of a read-write file on a read-only branch.\n");
 		bstart = fbstart(file);
 		bend = fbend(file);
 
-		PASSERT(file);
-		PASSERT(file->f_dentry);
-		PASSERT(file->f_dentry->d_inode);
-		PASSERT(file->f_dentry->d_inode);
-		ASSERT(S_ISREG(file->f_dentry->d_inode->i_mode));
+		BUG_ON(!S_ISREG(file->f_dentry->d_inode->i_mode));
 
 		for (bindex = bstart - 1; bindex >= 0; bindex--) {
 			if (!d_deleted(file->f_dentry)) {
@@ -349,9 +328,7 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 			bend = fbend(file);
 			for (bindex = bstart; bindex <= bend; bindex++) {
 				if (ftohf_index(file, bindex)) {
-					unionfs_read_lock(dentry->d_sb);
 					branchput(dentry->d_sb, bindex);
-					unionfs_read_unlock(dentry->d_sb);
 					fput(ftohf_index(file, bindex));
 					set_ftohf_index(file, bindex, NULL);
 				}
@@ -361,8 +338,9 @@ int unionfs_file_revalidate(struct file *file, int willwrite)
 	}
 
       out:
-	fist_print_dentry("file revalidate out", dentry);
+	print_dentry("file revalidate out", dentry);
 	unlock_dentry(dentry);
+	unionfs_read_unlock(dentry->d_sb);
 	print_exit_status(err);
 	return err;
 }
@@ -376,39 +354,33 @@ int unionfs_open(struct inode *inode, struct file *file)
 	struct dentry *dentry = NULL;
 	int bindex = 0, bstart = 0, bend = 0;
 	int locked = 0;
+	int size;
 
 	print_entry_location();
 
-	ftopd_lhs(file) =
-	    KMALLOC(sizeof(struct unionfs_file_info), GFP_UNIONFS);
+	ftopd_lhs(file) = KZALLOC(sizeof(struct unionfs_file_info), GFP_KERNEL);
 	if (!ftopd(file)) {
 		err = -ENOMEM;
 		goto out;
 	}
-	memset(ftopd(file), 0, sizeof(struct unionfs_file_info));
 	fbstart(file) = -1;
 	fbend(file) = -1;
 	atomic_set(&ftopd(file)->ufi_generation,
 		   atomic_read(&itopd(inode)->uii_generation));
-	if (sbmax(inode->i_sb) > UNIONFS_INLINE_OBJECTS) {
-		int size =
-		    sizeof(struct file *) * (sbmax(inode->i_sb) -
-					     UNIONFS_INLINE_OBJECTS);
-		ftohf_ptr(file) = KMALLOC(size, GFP_UNIONFS);
-		if (!ftohf_ptr(file)) {
-			err = -ENOMEM;
-			goto out;
-		}
-		memset(ftohf_ptr(file), 0, size);
+
+	size = sizeof(struct file *) * sbmax(inode->i_sb);
+	ftohf_ptr(file) = KZALLOC(size, GFP_KERNEL);
+	if (!ftohf_ptr(file)) {
+		err = -ENOMEM;
+		goto out;
 	}
-	memset(ftohf_inline(file), 0,
-	       sizeof(struct file *) * UNIONFS_INLINE_OBJECTS);
 
 	hidden_flags = file->f_flags;
 
 	dentry = file->f_dentry;
-	PASSERT(dentry);
+	dprint(PRINT_DEBUG, "dentry to open is %p\n", dentry);
 	lock_dentry(dentry);
+	unionfs_read_lock(inode->i_sb);
 	locked = 1;
 
 	bstart = fbstart(file) = dbstart(dentry);
@@ -445,7 +417,6 @@ int unionfs_open(struct inode *inode, struct file *file)
 	} else {
 		/* open a file */
 		hidden_dentry = dtohd(dentry);
-		PASSERT(hidden_dentry);
 
 		/* check for the permission for hidden file.  If the error is COPYUP_ERR,
 		 * copyup the file.
@@ -494,24 +465,28 @@ int unionfs_open(struct inode *inode, struct file *file)
       out:
 	/* freeing the allocated resources, and fput the opened files */
 	if (err < 0 && ftopd(file)) {
+		if (!locked)
+			unionfs_read_lock(file->f_dentry->d_sb);
 		for (bindex = bstart; bindex <= bend; bindex++) {
 			hidden_file = ftohf_index(file, bindex);
 			if (hidden_file) {
-				unionfs_read_lock(file->f_dentry->d_sb);
 				branchput(file->f_dentry->d_sb, bindex);
-				unionfs_read_unlock(file->f_dentry->d_sb);
 				/* fput calls dput for hidden_dentry */
 				fput(hidden_file);
 			}
 		}
+		if (!locked)
+			unionfs_read_unlock(file->f_dentry->d_sb);
 		KFREE(ftohf_ptr(file));
 		KFREE(ftopd(file));
 	}
 
-	fist_print_file("OUT: unionfs_open", file);
+	print_file("OUT: unionfs_open", file);
 
-	if (locked)
+	if (locked) {
 		unlock_dentry(dentry);
+		unionfs_read_unlock(inode->i_sb);
+	}
 	print_exit_status(err);
 	return err;
 }
@@ -525,7 +500,7 @@ int unionfs_file_release(struct inode *inode, struct file *file)
 
 	print_entry_location();
 
-	fist_checkinode(inode, "unionfs_release");
+	checkinode(inode, "unionfs_release");
 
 	/* fput all the hidden files */
 	fgen = atomic_read(&ftopd(file)->ufi_generation);
@@ -537,17 +512,19 @@ int unionfs_file_release(struct inode *inode, struct file *file)
 
 		if (hidden_file) {
 			fput(hidden_file);
+			unionfs_read_lock(inode->i_sb);
 			branchput_gen(fgen, inode->i_sb, bindex);
+			unionfs_read_unlock(inode->i_sb);
 		}
 	}
 	KFREE(ftohf_ptr(file));
 
 	if (ftopd(file)->rdstate) {
 		ftopd(file)->rdstate->uds_access = jiffies;
-		fist_dprint(1, "Saving rdstate with cookie %u [%d.%lld]\n",
-			    ftopd(file)->rdstate->uds_cookie,
-			    ftopd(file)->rdstate->uds_bindex,
-			    (long long)ftopd(file)->rdstate->uds_dirpos);
+		dprint(PRINT_DEBUG, "Saving rdstate with cookie %u [%d.%lld]\n",
+		       ftopd(file)->rdstate->uds_cookie,
+		       ftopd(file)->rdstate->uds_bindex,
+		       (long long)ftopd(file)->rdstate->uds_dirpos);
 		spin_lock(&itopd(inode)->uii_rdlock);
 		itopd(inode)->uii_rdcount++;
 		list_add_tail(&ftopd(file)->rdstate->uds_cache,
@@ -558,7 +535,7 @@ int unionfs_file_release(struct inode *inode, struct file *file)
 	}
 	KFREE(ftopd(file));
 
-	fist_checkinode(inode, "post unionfs_release");
+	checkinode(inode, "post unionfs_release");
 
 	print_exit_status(err);
 	return err;
@@ -582,8 +559,8 @@ long unionfs_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
 			err = -EACCES;
 			goto out;
 		}
-		val = fist_get_debug_value();
-		err = put_user(val, (int *)arg);
+		val = get_debug_mask();
+		err = put_user(val, (int __user *)arg);
 		break;
 
 	case FIST_IOCTL_SET_DEBUG_VALUE:
@@ -591,15 +568,15 @@ long unionfs_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
 			err = -EACCES;
 			goto out;
 		}
-		err = get_user(val, (int *)arg);
+		err = get_user(val, (int __user *)arg);
 		if (err)
 			break;
-		fist_dprint(6, "IOCTL SET: got arg %d\n", val);
-		if (val < 0 || val > 20) {
+		dprint(PRINT_DEBUG, "IOCTL SET: got arg %d\n", val);
+		if (val < 0 || val > PRINT_MAX) {
 			err = -EINVAL;
 			break;
 		}
-		fist_set_debug_value(val);
+		set_debug_mask(val);
 		break;
 
 		/* add non-debugging fist ioctl's here */
