@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York*
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,13 +16,11 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: copyup.c,v 1.48 2005/10/24 18:52:17 cwright Exp $
+ *  $Id: copyup.c,v 1.72 2006/06/01 03:11:02 jsipek Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 
-#if defined(UNIONFS_XATTR)
 /*Not Working Yet*/
 static int copyup_xattrs(struct dentry *old_hidden_dentry,
 			 struct dentry *new_hidden_dentry)
@@ -34,23 +33,8 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 
 	print_entry_location();
 
-	PASSERT(old_hidden_dentry);
-	PASSERT(old_hidden_dentry->d_inode);
-	PASSERT(old_hidden_dentry->d_inode->i_op);
-	PASSERT(new_hidden_dentry);
-	PASSERT(new_hidden_dentry->d_inode);
-	PASSERT(new_hidden_dentry->d_inode->i_op);
+	list_size = vfs_listxattr(old_hidden_dentry, NULL, 0);
 
-	if (!old_hidden_dentry->d_inode->i_op->getxattr ||
-	    !old_hidden_dentry->d_inode->i_op->listxattr ||
-	    !new_hidden_dentry->d_inode->i_op->setxattr) {
-		err = -ENOTSUPP;
-		goto out;
-	}
-
-	list_size =
-	    old_hidden_dentry->d_inode->i_op->listxattr(old_hidden_dentry, NULL,
-							0);
 	if (list_size <= 0) {
 		err = list_size;
 		goto out;
@@ -61,9 +45,7 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 		err = PTR_ERR(name_list);
 		goto out;
 	}
-	list_size =
-	    old_hidden_dentry->d_inode->i_op->listxattr(old_hidden_dentry,
-							name_list, list_size);
+	list_size = vfs_listxattr(old_hidden_dentry, name_list, list_size);
 	attr_value = xattr_alloc(XATTR_SIZE_MAX, XATTR_SIZE_MAX);
 	if (!attr_value || IS_ERR(attr_value)) {
 		err = PTR_ERR(name_list);
@@ -72,16 +54,15 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 	name_list_orig = name_list;
 	while (*name_list) {
 		ssize_t size;
-		down(&old_hidden_dentry->d_inode->i_sem);
+		mutex_lock(&old_hidden_dentry->d_inode->i_mutex);
 		err = security_inode_getxattr(old_hidden_dentry, name_list);
 		if (err)
 			size = err;
 		else
 			size =
-			    old_hidden_dentry->d_inode->i_op->
-			    getxattr(old_hidden_dentry, name_list, attr_value,
-				     XATTR_SIZE_MAX);
-		up(&old_hidden_dentry->d_inode->i_sem);
+			    vfs_getxattr(old_hidden_dentry, name_list,
+					 attr_value, XATTR_SIZE_MAX);
+		mutex_unlock(&old_hidden_dentry->d_inode->i_mutex);
 		if (size < 0) {
 			err = size;
 			goto out;
@@ -92,7 +73,7 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 			goto out;
 		}
 
-		down(&new_hidden_dentry->d_inode->i_sem);
+		mutex_lock(&new_hidden_dentry->d_inode->i_mutex);
 
 		err =
 		    security_inode_setxattr(old_hidden_dentry, name_list,
@@ -109,7 +90,7 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 							     attr_value, size,
 							     0);
 		}
-		up(&new_hidden_dentry->d_inode->i_sem);
+		mutex_unlock(&new_hidden_dentry->d_inode->i_mutex);
 
 		if (err < 0)
 			goto out;
@@ -123,11 +104,10 @@ static int copyup_xattrs(struct dentry *old_hidden_dentry,
 	if (attr_value)
 		xattr_free(attr_value, XATTR_SIZE_MAX);
 	/* It is no big deal if this fails, we just roll with the punches. */
-	if (err == -ENOTSUPP)
+	if (err == -ENOTSUPP || err == -EOPNOTSUPP)
 		err = 0;
 	return err;
 }
-#endif
 
 /* Determine the mode based on the copyup flags, and the existing dentry. */
 static int copyup_permissions(struct super_block *sb,
@@ -139,31 +119,16 @@ static int copyup_permissions(struct super_block *sb,
 
 	print_entry_location();
 
-	newattrs.ia_valid = ATTR_CTIME;
-	if (IS_SET(sb, COPYUP_FS_MOUNTER)) {
-		/* f/s mounter */
-		newattrs.ia_mode = stopd(sb)->copyupmode;
-		newattrs.ia_valid |= ATTR_FORCE | ATTR_MODE;
-		if (stopd(sb)->copyupuid != -1) {
-			newattrs.ia_uid = stopd(sb)->copyupuid;
-			newattrs.ia_valid |= ATTR_UID;
-		}
-		if (stopd(sb)->copyupgid != -1) {
-			newattrs.ia_gid = stopd(sb)->copyupgid;
-			newattrs.ia_valid |= ATTR_GID;
-		}
-	} else if (IS_SET(sb, COPYUP_CURRENT_USER)) {
-		/* current file permission */
-		newattrs.ia_mode = ~current->fs->umask & S_IRWXUGO;
-		newattrs.ia_valid |= ATTR_FORCE | ATTR_MODE;
-	} else {
-		/* original mode of old file */
-		newattrs.ia_mode = old_hidden_dentry->d_inode->i_mode;
-		newattrs.ia_gid = old_hidden_dentry->d_inode->i_gid;
-		newattrs.ia_uid = old_hidden_dentry->d_inode->i_uid;
-		newattrs.ia_valid |=
-		    ATTR_FORCE | ATTR_GID | ATTR_UID | ATTR_MODE;
-	}
+	newattrs.ia_atime = old_hidden_dentry->d_inode->i_atime;
+	newattrs.ia_mtime = old_hidden_dentry->d_inode->i_mtime;
+	newattrs.ia_ctime = old_hidden_dentry->d_inode->i_ctime;
+	newattrs.ia_valid = ATTR_CTIME | ATTR_ATIME | ATTR_MTIME |
+	    ATTR_ATIME_SET | ATTR_MTIME_SET;
+	/* original mode of old file */
+	newattrs.ia_mode = old_hidden_dentry->d_inode->i_mode;
+	newattrs.ia_gid = old_hidden_dentry->d_inode->i_gid;
+	newattrs.ia_uid = old_hidden_dentry->d_inode->i_uid;
+	newattrs.ia_valid |= ATTR_FORCE | ATTR_GID | ATTR_UID | ATTR_MODE;
 	if (newattrs.ia_valid & ATTR_MODE) {
 		newattrs.ia_mode =
 		    (newattrs.ia_mode & S_IALLUGO) | (old_hidden_dentry->
@@ -205,7 +170,7 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 	int old_bstart;
 	int old_bend;
 	int size = len;
-	struct dentry *new_hidden_parent_dentry;
+	struct dentry *new_hidden_parent_dentry = NULL;
 	mm_segment_t oldfs;
 	char *symbuf = NULL;
 	uid_t saved_uid = current->fsuid;
@@ -213,44 +178,39 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 
 	print_entry_location();
 	verify_locked(dentry);
-	fist_print_dentry("IN: copyup_named_dentry", dentry);
+	print_dentry("IN: copyup_named_dentry", dentry);
 
 	old_bindex = bstart;
 	old_bstart = dbstart(dentry);
 	old_bend = dbend(dentry);
 
-	ASSERT(new_bindex >= 0);
-	ASSERT(new_bindex < old_bindex);
-	PASSERT(dir);
-	PASSERT(dentry);
+	BUG_ON(new_bindex < 0);
+	BUG_ON(new_bindex >= old_bindex);
 
 	sb = dir->i_sb;
 
+	unionfs_read_lock(sb);
+
 	if ((err = is_robranch_super(sb, new_bindex)))
 		goto out;
 
 	/* Create the directory structure above this dentry. */
 	new_hidden_dentry = create_parents_named(dir, dentry, name, new_bindex);
-	PASSERT(new_hidden_dentry);
 	if (IS_ERR(new_hidden_dentry)) {
 		err = PTR_ERR(new_hidden_dentry);
 		goto out;
 	}
 
-	fist_print_generic_dentry("Copyup Object", new_hidden_dentry);
+	print_dentry("Copyup Object", new_hidden_dentry);
 
 	/* Now we actually create the object. */
 	old_hidden_dentry = dtohd_index(dentry, old_bindex);
-	PASSERT(old_hidden_dentry);
-	PASSERT(old_hidden_dentry->d_inode);
 	DGET(old_hidden_dentry);
 
 	/* For symlinks, we must read the link before we lock the directory. */
 	if (S_ISLNK(old_hidden_dentry->d_inode->i_mode)) {
-		PASSERT(old_hidden_dentry->d_inode->i_op);
-		PASSERT(old_hidden_dentry->d_inode->i_op->readlink);
 
-		symbuf = KMALLOC(PATH_MAX, GFP_UNIONFS);
+		symbuf = KMALLOC(PATH_MAX, GFP_KERNEL);
 		if (!symbuf) {
 			err = -ENOMEM;
 			goto copyup_readlink_err;
@@ -260,7 +220,8 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 		set_fs(KERNEL_DS);
 		err =
 		    old_hidden_dentry->d_inode->i_op->
-		    readlink(old_hidden_dentry, symbuf, PATH_MAX);
+		    readlink(old_hidden_dentry, (char __user *)symbuf,
+			     PATH_MAX);
 		set_fs(oldfs);
 		if (err < 0)
 			goto copyup_readlink_err;
@@ -289,14 +250,12 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 		err = vfs_create(new_hidden_parent_dentry->d_inode,
 				 new_hidden_dentry, S_IRWXU, NULL);
 	} else {
-		char diemsg[100];
-		snprintf(diemsg, sizeof(diemsg), "Unknown inode type %d\n",
-			 old_hidden_dentry->d_inode->i_mode);
-		FISTBUG(diemsg);
+		printk(KERN_ERR "Unknown inode type %d\n",
+		       old_hidden_dentry->d_inode->i_mode);
+		BUG();
 	}
 	current->fsuid = saved_uid;
 	current->fsgid = saved_gid;
-	unlock_dir(new_hidden_parent_dentry);
       copyup_readlink_err:
 	KFREE(symbuf);
 	if (err) {
@@ -305,9 +264,16 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 		set_dtohd_index(dentry, new_bindex, NULL);
 		set_dbstart(dentry, old_bstart);
 		set_dbend(dentry, old_bend);
-		goto out;
+		goto out_dir;
 	}
-
+#ifdef UNIONFS_IMAP
+	if (stopd(sb)->usi_persistent) {
+		err = write_uin(dentry->d_sb, dentry->d_inode->i_ino,
+				new_bindex, new_hidden_dentry->d_inode->i_ino);
+		if (err)
+			goto out_dir;
+	}
+#endif
 	/* We actually copyup the file here. */
 	if (S_ISREG(old_hidden_dentry->d_inode->i_mode)) {
 		mntget(stohiddenmnt_index(sb, old_bindex));
@@ -318,11 +284,11 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 				stohiddenmnt_index(sb, old_bindex), O_RDONLY);
 		if (IS_ERR(input_file)) {
 			err = PTR_ERR(input_file);
-			goto out;
+			goto out_dir;
 		}
 		if (!input_file->f_op || !input_file->f_op->read) {
 			err = -EINVAL;
-			goto out;
+			goto out_dir;
 		}
 
 		/* copy the new file */
@@ -335,18 +301,18 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 				stohiddenmnt_index(sb, new_bindex), O_WRONLY);
 		if (IS_ERR(output_file)) {
 			err = PTR_ERR(output_file);
-			goto out;
+			goto out_dir;
 		}
 		if (!output_file->f_op || !output_file->f_op->write) {
 			err = -EINVAL;
-			goto out;
+			goto out_dir;
 		}
 
 		/* allocating a buffer */
-		buf = (char *)KMALLOC(PAGE_SIZE, GFP_UNIONFS);
+		buf = (char *)KMALLOC(PAGE_SIZE, GFP_KERNEL);
 		if (!buf) {
 			err = -ENOMEM;
-			goto out;
+			goto out_dir;
 		}
 
 		/* now read PAGE_SIZE bytes from offset 0 in a loop */
@@ -355,6 +321,7 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 		input_file->f_pos = 0;
 		output_file->f_pos = 0;
 
+		err = 0;	// reset error just in case
 		set_fs(KERNEL_DS);
 		do {
 			if (len >= PAGE_SIZE)
@@ -365,7 +332,8 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 			len -= PAGE_SIZE;
 
 			read_bytes =
-			    input_file->f_op->read(input_file, buf, size,
+			    input_file->f_op->read(input_file,
+						   (char __user *)buf, size,
 						   &input_file->f_pos);
 			if (read_bytes <= 0) {
 				err = read_bytes;
@@ -373,32 +341,59 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 			}
 
 			write_bytes =
-			    output_file->f_op->write(output_file, buf,
+			    output_file->f_op->write(output_file,
+						     (char __user *)buf,
 						     read_bytes,
 						     &output_file->f_pos);
 			if (write_bytes < 0 || (write_bytes < read_bytes)) {
-				err = -EIO;
+				err = write_bytes;
 				break;
 			}
 		} while ((read_bytes > 0) && (len > 0));
 		set_fs(old_fs);
 		KFREE(buf);
+#ifdef UNIONFS_MMAP
+		/* SP: Now that we copied up the file, have to sync its data
+		 * as otherwise when we do a read_cache_page(), we'll possibly
+		 * read crap.
+		 *
+		 * another posisble solution would be in the address op code
+		 * would be to check the "lower" page to see if its dirty,
+		 * and if it's dirty, use it directl 
+		 */
+		if (!err) {
+			err =
+			    output_file->f_op->fsync(output_file,
+						     new_hidden_dentry, 0);
+		}
+#endif
+		if (err) {
+			/* copyup failed, because we ran out of space or quota,
+			 * or something else happened so let's unlink; we don't
+			 * really care about the return value of vfs_unlink */
+			vfs_unlink(new_hidden_parent_dentry->d_inode,
+				   new_hidden_dentry);
+
+			goto out_dir;
+		}
 	}
 
 	/* Set permissions. */
 	if ((err =
 	     copyup_permissions(sb, old_hidden_dentry, new_hidden_dentry)))
-		goto out;
+		goto out_dir;
 	/* Selinux uses extended attributes for permissions. */
-#if defined(UNIONFS_XATTR)
 	if ((err = copyup_xattrs(old_hidden_dentry, new_hidden_dentry)))
-		goto out;
-#endif
+		goto out_dir;
 
 	/* do not allow files getting deleted to be reinterposed */
 	if (!d_deleted(dentry))
 		unionfs_reinterpose(dentry);
 
+      out_dir:
+	if (new_hidden_parent_dentry)
+		unlock_dir(new_hidden_parent_dentry);
+
       out:
 	if (input_file && !IS_ERR(input_file)) {
 		fput(input_file);
@@ -417,15 +412,18 @@ int copyup_named_dentry(struct inode *dir, struct dentry *dentry,
 		if (copyup_file && !err) {
 			*copyup_file = output_file;
 		} else {
-			if (!err) {
+			/* close the file if there was no error, or if we ran
+			 * out of space in which case we unlinked the file */
+			if (!IS_ERR(output_file))
 				fput(output_file);
-			}
 			branchput(sb, got_branch_output);
 		}
 	}
 
-	fist_print_dentry("OUT: copyup_dentry", dentry);
-	fist_print_inode("OUT: copyup_dentry", dentry->d_inode);
+	unionfs_read_unlock(sb);
+
+	print_dentry("OUT: copyup_dentry", dentry);
+	print_inode("OUT: copyup_dentry", dentry->d_inode);
 
 	print_exit_status(err);
 	return err;
@@ -515,7 +513,10 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 	int old_bend;
 	struct dentry **path = NULL;
 	struct dentry **tmp_path;
-
+	struct super_block *sb;
+#ifdef UNIONFS_IMAP
+	int persistent;
+#endif
 	print_entry_location();
 
 	verify_locked(dentry);
@@ -529,14 +530,16 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 		goto out;
 	}
 
-	fist_print_dentry("IN: create_parents_named", dentry);
-	fist_dprint(8, "name = %s\n", name);
+	print_dentry("IN: create_parents_named", dentry);
+	dprint(PRINT_DEBUG, "name = %s\n", name);
 
 	old_bstart = dbstart(dentry);
 	old_bend = dbend(dentry);
 
-	path = (struct dentry **)KMALLOC(kmalloc_size, GFP_KERNEL);
-	memset(path, 0, kmalloc_size);
+	hidden_dentry = ERR_PTR(-ENOMEM);
+	path = (struct dentry **)KZALLOC(kmalloc_size, GFP_KERNEL);
+	if (!path)
+		goto out;
 
 	/* assume the negative dentry of unionfs as the parent dentry */
 	parent_dentry = dentry;
@@ -568,12 +571,11 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 			num_dentry = kmalloc_size / sizeof(struct dentry *);
 
 			tmp_path =
-			    (struct dentry **)KMALLOC(kmalloc_size, GFP_KERNEL);
+			    (struct dentry **)KZALLOC(kmalloc_size, GFP_KERNEL);
 			if (!tmp_path) {
 				hidden_dentry = ERR_PTR(-ENOMEM);
 				goto out;
 			}
-			memset(tmp_path, 0, kmalloc_size);
 			memcpy(tmp_path, path, old_kmalloc_size);
 			KFREE(path);
 			path = tmp_path;
@@ -583,18 +585,16 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 	} while (!hidden_parent_dentry);
 	count--;
 
+	sb = dentry->d_sb;
+#ifdef UNIONFS_IMAP
+	persistent = stopd(sb)->usi_persistent;
+#endif
 	/* This is basically while(child_dentry != dentry).  This loop is
 	 * horrible to follow and should be replaced with cleaner code. */
 	while (1) {
-		PASSERT(child_dentry);
-		PASSERT(parent_dentry);
-		PASSERT(parent_dentry->d_inode);
-
 		// get hidden parent dir in the current branch
 		hidden_parent_dentry = dtohd_index(parent_dentry, bindex);
 		unlock_dentry(parent_dentry);
-		PASSERT(hidden_parent_dentry);
-		PASSERT(hidden_parent_dentry->d_inode);
 
 		// init the values to lookup
 		childname = child_dentry->d_name.name;
@@ -681,22 +681,29 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 					hidden_dentry, S_IRWXUGO);
 			current->fsuid = saved_uid;
 			current->fsgid = saved_gid;
+			if (!err)
+				err = copyup_permissions
+				    (dir->i_sb, child_dentry, hidden_dentry);
 			unlock_dir(hidden_parent_dentry);
-			if (err || !hidden_dentry->d_inode) {
-				DPUT(hidden_dentry);
-				hidden_dentry = ERR_PTR(err);
-				goto out;
-			}
-			err =
-			    copyup_permissions(dir->i_sb, child_dentry,
-					       hidden_dentry);
 			if (err) {
 				DPUT(hidden_dentry);
 				hidden_dentry = ERR_PTR(err);
 				goto out;
 			}
+#ifdef UNIONFS_IMAP
+			if (persistent) {
+				err = write_uin
+				    (sb, child_dentry->d_inode->i_ino,
+				     bindex, hidden_dentry->d_inode->i_ino);
+				if (err) {
+					DPUT(hidden_dentry);
+					hidden_dentry = ERR_PTR(err);
+					goto out;
+				}
+			}
+#endif
 			set_itohi_index(child_dentry->d_inode, bindex,
-					igrab(hidden_dentry->d_inode));
+					IGRAB(hidden_dentry->d_inode));
 			if (ibstart(child_dentry->d_inode) > bindex)
 				ibstart(child_dentry->d_inode) = bindex;
 			if (ibend(child_dentry->d_inode) < bindex)
@@ -714,7 +721,7 @@ struct dentry *create_parents_named(struct inode *dir, struct dentry *dentry,
 	}
       out:
 	KFREE(path);
-	fist_print_dentry("OUT: create_parents_named", dentry);
+	print_dentry("OUT: create_parents_named", dentry);
 	print_exit_pointer(hidden_dentry);
 	return hidden_dentry;
 }
