@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,10 +16,9 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: dirhelper.c,v 1.18 2005/10/20 17:22:19 cwright Exp $
+ *  $Id: dirhelper.c,v 1.29 2006/02/21 08:36:24 jsipek Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 
 /* Delete all of the whiteouts in a given directory for rmdir. */
@@ -29,7 +29,10 @@ int delete_whiteouts(struct dentry *dentry, int bindex,
 	struct dentry *hidden_dir_dentry = NULL;
 	struct dentry *hidden_dentry;
 	struct super_block *sb;
-	char *name = NULL;
+	char *name = NULL, *p;
+	struct inode *hidden_dir;
+	struct superio sio;
+	int do_superio;
 
 	int i;
 	struct list_head *pos;
@@ -38,29 +41,34 @@ int delete_whiteouts(struct dentry *dentry, int bindex,
 	print_entry_location();
 
 	sb = dentry->d_sb;
+	unionfs_read_lock(sb);
 
-	ASSERT(S_ISDIR(dentry->d_inode->i_mode));
-
-	/* Make sure all of the hidden_dentries are filled in. */
-	if ((err = unionfs_partial_lookup(dentry)))
+	BUG_ON(!S_ISDIR(dentry->d_inode->i_mode));
+	BUG_ON(bindex < dbstart(dentry));
+	BUG_ON(bindex > dbend(dentry));
+	err = is_robranch_super(sb, bindex);
+	if (err)
 		goto out;
 
-	ASSERT(bindex >= dbstart(dentry));
-	ASSERT(bindex <= dbend(dentry));
-
 	/* Find out hidden parent dentry */
 	hidden_dir_dentry = dtohd_index(dentry, bindex);
-	PASSERT(hidden_dir_dentry);
-	PASSERT(hidden_dir_dentry->d_inode);
-	ASSERT(S_ISDIR(hidden_dir_dentry->d_inode->i_mode));
+	BUG_ON(!S_ISDIR(hidden_dir_dentry->d_inode->i_mode));
+	hidden_dir = hidden_dir_dentry->d_inode;
+	BUG_ON(!S_ISDIR(hidden_dir->i_mode));
 
-	name = (char *)__get_free_page(GFP_UNIONFS);
-	if (!name) {
-		err = -ENOMEM;
+	err = -ENOMEM;
+	name = __getname();
+	if (!name)
 		goto out;
-	}
+	strcpy(name, WHPFX);
+	p = name + WHLEN;
 
-	for (i = 0; i < namelist->uds_size; i++) {
+	err = 0;
+	mutex_lock(&hidden_dir->i_mutex);
+	do_superio = permission(hidden_dir, MAY_WRITE | MAY_EXEC, NULL);
+	if (do_superio)
+		superio_store(&sio);
+	for (i = 0; !err && i < namelist->uds_size; i++) {
 		list_for_each(pos, &namelist->uds_list[i]) {
 			cursor =
 			    list_entry(pos, struct filldir_node, file_list);
@@ -70,40 +78,32 @@ int delete_whiteouts(struct dentry *dentry, int bindex,
 			if (!cursor->whiteout)
 				continue;
 
-			strcpy(name, ".wh.");
-			strncpy(name + 4, cursor->name, PAGE_SIZE - 4);
-
+			strcpy(p, cursor->name);
 			hidden_dentry =
 			    LOOKUP_ONE_LEN(name, hidden_dir_dentry,
-					   cursor->namelen + 4);
+					   cursor->namelen + WHLEN);
 			if (IS_ERR(hidden_dentry)) {
 				err = PTR_ERR(hidden_dentry);
-				goto out;
+				break;
 			}
-			if (!hidden_dentry->d_inode) {
-				DPUT(hidden_dentry);
-				continue;
-			}
-
-			down(&hidden_dir_dentry->d_inode->i_sem);
-			err =
-			    vfs_unlink(hidden_dir_dentry->d_inode,
-				       hidden_dentry);
-			up(&hidden_dir_dentry->d_inode->i_sem);
+			if (hidden_dentry->d_inode)
+				err = vfs_unlink(hidden_dir, hidden_dentry);
 			DPUT(hidden_dentry);
-
-			if (err && !IS_COPYUP_ERR(err))
-				goto out;
+			if (err)
+				break;
 		}
 	}
+	mutex_unlock(&hidden_dir->i_mutex);
+	if (do_superio)
+		superio_revert(&sio);
+
+	__putname(name);
 
-      out:
 	/* After all of the removals, we should copy the attributes once. */
 	fist_copy_attr_times(dentry->d_inode, hidden_dir_dentry->d_inode);
-	dentry->d_inode->i_nlink = get_nlinks(dentry->d_inode);
 
-	if (name)
-		free_page((unsigned long)name);
+      out:
+	unionfs_read_unlock(sb);
 	print_exit_status(err);
 	return err;
 }
@@ -136,9 +136,9 @@ static int readdir_util_callback(void *dirent, const char *name, int namelen,
 	    && (namelen == 1 || (name[1] == '.' && namelen == 2)))
 		goto out;
 
-	if ((namelen > 4) && !strncmp(name, ".wh.", 4)) {
-		namelen -= 4;
-		name += 4;
+	if ((namelen > WHLEN) && !strncmp(name, WHPFX, WHLEN)) {
+		namelen -= WHLEN;
+		name += WHLEN;
 		whiteout = 1;
 	}
 
@@ -175,7 +175,9 @@ int check_empty(struct dentry *dentry, struct unionfs_dir_state **namelist)
 
 	sb = dentry->d_sb;
 
-	ASSERT(S_ISDIR(dentry->d_inode->i_mode));
+	unionfs_read_lock(sb);
+
+	BUG_ON(!S_ISDIR(dentry->d_inode->i_mode));
 
 	if ((err = unionfs_partial_lookup(dentry)))
 		goto out;
@@ -186,7 +188,7 @@ int check_empty(struct dentry *dentry, struct unionfs_dir_state **namelist)
 	if (0 <= bopaque && bopaque < bend)
 		bend = bopaque;
 
-	buf = KMALLOC(sizeof(struct unionfs_rdutil_callback), GFP_UNIONFS);
+	buf = KMALLOC(sizeof(struct unionfs_rdutil_callback), GFP_KERNEL);
 	if (!buf) {
 		err = -ENOMEM;
 		goto out;
@@ -248,6 +250,8 @@ int check_empty(struct dentry *dentry, struct unionfs_dir_state **namelist)
 		KFREE(buf);
 	}
 
+	unionfs_read_unlock(sb);
+
 	print_exit_status(err);
 	return err;
 }
