@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,10 +16,9 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: lookup.c,v 1.27 2005/09/18 05:02:56 jsipek Exp $
+ *  $Id: lookup.c,v 1.46 2006/06/01 03:11:03 jsipek Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 
 static int is_opaque_dir(struct dentry *dentry, int bindex);
@@ -44,7 +44,6 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 	int namelen;
 
 	print_entry("mode = %d", lookupmode);
-	PASSERT(dentry);
 
 	/* We should already have a lock on this dentry in the case of a
 	 * partial lookup, or a revalidation. Otherwise it is returned from
@@ -53,26 +52,28 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 	    || lookupmode == INTERPOSE_REVAL_NEG) {
 		verify_locked(dentry);
 	} else {
-		ASSERT(dtopd_nocheck(dentry) == NULL);
+		BUG_ON(dtopd_nocheck(dentry) != NULL);
 		locked_child = 1;
 	}
 	if (lookupmode != INTERPOSE_PARTIAL)
 		if ((err = new_dentry_private_data(dentry)))
 			goto out;
+	/* must initialize dentry operations */
+	dentry->d_op = &unionfs_dops;
+
 	parent_dentry = GET_PARENT(dentry);
 	/* We never partial lookup the root directory. */
 	if (parent_dentry != dentry) {
 		lock_dentry(parent_dentry);
 		locked_parent = 1;
-		PASSERT(parent_dentry);
 	} else {
 		DPUT(parent_dentry);
 		parent_dentry = NULL;
 		goto out;
 	}
 
-	fist_print_dentry("IN unionfs_lookup (parent)", parent_dentry);
-	fist_print_dentry("IN unionfs_lookup (child)", dentry);
+	print_dentry("IN unionfs_lookup (parent)", parent_dentry);
+	print_dentry("IN unionfs_lookup (child)", dentry);
 
 	name = dentry->d_name.name;
 	namelen = dentry->d_name.len;
@@ -83,14 +84,11 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 		goto out_free;
 	}
 
-	/* must initialize dentry operations */
-	dentry->d_op = &unionfs_dops;
-
 	/* Now start the actual lookup procedure. */
 	bstart = dbstart(parent_dentry);
 	bend = dbend(parent_dentry);
 	bopaque = dbopaque(parent_dentry);
-	ASSERT(bstart >= 0);
+	BUG_ON(bstart < 0);
 
 	/* It would be ideal if we could convert partial lookups to only have
 	 * to do this work when they really need to.  It could probably improve
@@ -101,12 +99,12 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 			bend = bopaque;
 	}
 
-	fist_dprint(8, "bstart = %d, bend = %d\n", bstart, bend);
+	dprint(PRINT_DEBUG, "bstart = %d, bend = %d\n", bstart, bend);
 	for (bindex = bstart; bindex <= bend; bindex++) {
 		hidden_dentry = dtohd_index(dentry, bindex);
 		if (lookupmode == INTERPOSE_PARTIAL && hidden_dentry)
 			continue;
-		ASSERT(hidden_dentry == NULL);
+		BUG_ON(hidden_dentry != NULL);
 
 		hidden_dir_dentry = dtohd_index(parent_dentry, bindex);
 
@@ -114,43 +112,42 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 		if (!(hidden_dir_dentry && hidden_dir_dentry->d_inode))
 			continue;
 
-		PASSERT(hidden_dir_dentry->d_inode);
 		/* also skip it if the parent isn't a directory. */
 		if (!S_ISDIR(hidden_dir_dentry->d_inode->i_mode))
 			continue;
 
 		/* Reuse the whiteout name because its value doesn't change. */
 		if (!whname) {
-			whname = KMALLOC(namelen + 5, GFP_UNIONFS);
-			if (!whname) {
-				err = -ENOMEM;
+			whname = alloc_whname(name, namelen);
+			if (IS_ERR(whname)) {
+				err = PTR_ERR(whname);
 				goto out_free;
 			}
-			strcpy(whname, ".wh.");
-			strncat(whname, name, namelen);
-			whname[4 + namelen] = '\0';
 		}
 
 		/* check if whiteout exists in this branch: lookup .wh.foo */
 		wh_hidden_dentry = LOOKUP_ONE_LEN(whname, hidden_dir_dentry,
-						  namelen + 4);
+						  namelen + WHLEN);
 		if (IS_ERR(wh_hidden_dentry)) {
+			DPUT(first_hidden_dentry);
 			err = PTR_ERR(wh_hidden_dentry);
 			goto out_free;
 		}
 
 		if (wh_hidden_dentry->d_inode) {
-			DPUT(wh_hidden_dentry);
 			/* We found a whiteout so lets give up. */
-			fist_dprint(8, "whiteout found in %d\n", bindex);
+			dprint(PRINT_DEBUG, "whiteout found in %d\n", bindex);
 			if (S_ISREG(wh_hidden_dentry->d_inode->i_mode)) {
 				set_dbend(dentry, bindex);
 				set_dbopaque(dentry, bindex);
+				DPUT(wh_hidden_dentry);
 				break;
 			}
 			err = -EIO;
 			printk(KERN_NOTICE "EIO: Invalid whiteout entry type"
 			       " %d.\n", wh_hidden_dentry->d_inode->i_mode);
+			DPUT(wh_hidden_dentry);
+			DPUT(first_hidden_dentry);
 			goto out_free;
 		}
 
@@ -160,8 +157,9 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 		/* Now do regular lookup; lookup foo */
 		hidden_dentry = LOOKUP_ONE_LEN(name, hidden_dir_dentry,
 					       namelen);
-		fist_print_generic_dentry("hidden result", hidden_dentry);
+		print_dentry("hidden result", hidden_dentry);
 		if (IS_ERR(hidden_dentry)) {
+			DPUT(first_hidden_dentry);
 			err = PTR_ERR(hidden_dentry);
 			goto out_free;
 		}
@@ -198,14 +196,13 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 			if (dentry_count == 1)
 				goto out_positive;
 			/* This can only happen with mixed D-*-F-* */
-			PASSERT(dtohd(dentry));
-			PASSERT(dtohd(dentry)->d_inode);
-			ASSERT(S_ISDIR(dtohd(dentry)->d_inode->i_mode));
+			BUG_ON(!S_ISDIR(dtohd(dentry)->d_inode->i_mode));
 			continue;
 		}
 
 		opaque = is_opaque_dir(dentry, bindex);
 		if (opaque < 0) {
+			DPUT(first_hidden_dentry);
 			err = opaque;
 			goto out_free;
 		}
@@ -228,14 +225,12 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 	/* If we've only got negative dentries, then use the leftmost one. */
 	if (lookupmode == INTERPOSE_REVAL) {
 		if (dentry->d_inode) {
-			PASSERT(itopd(dentry->d_inode));
 			itopd(dentry->d_inode)->uii_stale = 1;
 		}
 		goto out;
 	}
 	/* This should only happen if we found a whiteout. */
 	if (first_dentry_offset == -1) {
-		PASSERT(hidden_dir_dentry);
 		first_hidden_dentry = LOOKUP_ONE_LEN(name, hidden_dir_dentry,
 						     namelen);
 		first_dentry_offset = bindex;
@@ -249,14 +244,14 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 	set_dbend(dentry, first_dentry_offset);
 
 	if (lookupmode == INTERPOSE_REVAL_NEG)
-		ASSERT(dentry->d_inode == NULL);
+		BUG_ON(dentry->d_inode != NULL);
 	else
 		d_add(dentry, NULL);
 	goto out;
 
 /* This part of the code is for positive dentries. */
       out_positive:
-	ASSERT(dentry_count > 0);
+	BUG_ON(dentry_count <= 0);
 
 	/* If we're holding onto the first negative dentry throw it out. */
 	DPUT(first_hidden_dentry);
@@ -281,8 +276,8 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 	if (err)
 		goto out_drop;
 
-	fist_checkinode(dentry->d_inode, "unionfs_lookup OUT: child");
-	fist_checkinode(parent_dentry->d_inode, "unionfs_lookup OUT: dir");
+	checkinode(dentry->d_inode, "unionfs_lookup OUT: child");
+	checkinode(parent_dentry->d_inode, "unionfs_lookup OUT: dir");
 	goto out;
 
       out_drop:
@@ -303,13 +298,13 @@ struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode)
 
       out:
 	if (!err && dtopd(dentry)) {
-		ASSERT(dbend(dentry) <= dtopd(dentry)->udi_bcount);
-		ASSERT(dbend(dentry) <= sbmax(dentry->d_sb));
-		ASSERT(dbstart(dentry) >= 0);
+		BUG_ON(dbend(dentry) > dtopd(dentry)->udi_bcount);
+		BUG_ON(dbend(dentry) > sbmax(dentry->d_sb));
+		BUG_ON(dbstart(dentry) < 0);
 	}
 	KFREE(whname);
-	fist_print_dentry("OUT unionfs_lookup (parent)", parent_dentry);
-	fist_print_dentry("OUT unionfs_lookup (child)", dentry);
+	print_dentry("OUT unionfs_lookup (parent)", parent_dentry);
+	print_dentry("OUT unionfs_lookup (child)", dentry);
 	if (locked_parent)
 		unlock_dentry(parent_dentry);
 	DPUT(parent_dentry);
@@ -325,8 +320,13 @@ int unionfs_partial_lookup(struct dentry *dentry)
 	struct dentry *tmp;
 
 	tmp = unionfs_lookup_backend(dentry, INTERPOSE_PARTIAL);
-
-	return PTR_ERR(tmp);
+	if (!tmp)
+		return 0;
+	if (IS_ERR(tmp))
+		return PTR_ERR(tmp);
+	/* need to change the interface */
+	BUG_ON(tmp != dentry);
+	return -ENOSYS;
 }
 
 /* The rest of these are utility functions for lookup. */
@@ -335,20 +335,29 @@ static int is_opaque_dir(struct dentry *dentry, int bindex)
 	int err = 0;
 	struct dentry *hidden_dentry;
 	struct dentry *wh_hidden_dentry;
+	struct inode *hidden_inode;
+	struct superio sio;
+	int do_superio;
 
 	print_entry_location();
 
 	hidden_dentry = dtohd_index(dentry, bindex);
+	hidden_inode = hidden_dentry->d_inode;
 
-	PASSERT(hidden_dentry);
-	PASSERT(hidden_dentry->d_inode);
-	ASSERT(S_ISDIR(hidden_dentry->d_inode->i_mode));
+	BUG_ON(!S_ISDIR(hidden_inode->i_mode));
 
+	mutex_lock(&hidden_inode->i_mutex);
+	do_superio = permission(hidden_inode, MAY_EXEC, NULL);
+	if (do_superio)
+		superio_store(&sio);
 	wh_hidden_dentry = LOOKUP_ONE_LEN(UNIONFS_DIR_OPAQUE, hidden_dentry,
 					  sizeof(UNIONFS_DIR_OPAQUE) - 1);
+	if (do_superio)
+		superio_revert(&sio);
+	mutex_unlock(&hidden_inode->i_mutex);
 	if (IS_ERR(wh_hidden_dentry)) {
 		err = PTR_ERR(wh_hidden_dentry);
-		fist_dprint(1, "LOOKUP_ONE_LEN returned: %d\n", err);
+		dprint(PRINT_DEBUG, "LOOKUP_ONE_LEN returned: %d\n", err);
 		goto out;
 	}
 	if (wh_hidden_dentry->d_inode)
@@ -361,7 +370,7 @@ static int is_opaque_dir(struct dentry *dentry, int bindex)
 
 static int is_validname(const char *name)
 {
-	if (!strncmp(name, ".wh.", 4))
+	if (!strncmp(name, WHPFX, WHLEN))
 		return 0;
 	if (!strncmp(name, UNIONFS_DIR_OPAQUE_NAME,
 		     sizeof(UNIONFS_DIR_OPAQUE_NAME) - 1))
@@ -371,7 +380,7 @@ static int is_validname(const char *name)
 
 /* The dentry cache is just so we have properly sized dentries. */
 static kmem_cache_t *unionfs_dentry_cachep;
-int init_dentry_cache()
+int init_dentry_cache(void)
 {
 	unionfs_dentry_cachep =
 	    kmem_cache_create("unionfs_dentry",
@@ -383,7 +392,7 @@ int init_dentry_cache()
 	return 0;
 }
 
-void destroy_dentry_cache()
+void destroy_dentry_cache(void)
 {
 	if (!unionfs_dentry_cachep)
 		return;
@@ -405,9 +414,10 @@ int new_dentry_private_data(struct dentry *dentry)
 	int newsize;
 	int oldsize = 0;
 
+	spin_lock(&dentry->d_lock);
 	if (!dtopd_nocheck(dentry)) {
 		dtopd_lhs(dentry) = (struct unionfs_dentry_info *)
-		    kmem_cache_alloc(unionfs_dentry_cachep, SLAB_KERNEL);
+		    kmem_cache_alloc(unionfs_dentry_cachep, SLAB_ATOMIC);
 		if (!dtopd_nocheck(dentry))
 			goto out;
 		init_MUTEX_LOCKED(&dtopd_nocheck(dentry)->udi_sem);
@@ -416,10 +426,7 @@ int new_dentry_private_data(struct dentry *dentry)
 #endif
 		dtohd_ptr(dentry) = NULL;
 	} else {
-		if (dtopd(dentry)->udi_bcount > UNIONFS_INLINE_OBJECTS)
-			oldsize = sizeof(struct dentry *) *
-			    (dtopd(dentry)->udi_bcount -
-			     UNIONFS_INLINE_OBJECTS);
+		oldsize = sizeof(struct dentry *) * dtopd(dentry)->udi_bcount;
 	}
 
 	dtopd_nocheck(dentry)->udi_bstart = -1;
@@ -428,10 +435,7 @@ int new_dentry_private_data(struct dentry *dentry)
 	dtopd_nocheck(dentry)->udi_bcount = sbmax(dentry->d_sb);
 	atomic_set(&dtopd_nocheck(dentry)->udi_generation,
 		   atomic_read(&stopd(dentry->d_sb)->usi_generation));
-	newsize = (sbmax(dentry->d_sb) - UNIONFS_INLINE_OBJECTS)
-	    * sizeof(struct dentry *);
-	if (newsize < 0)
-		newsize = 0;
+	newsize = sizeof(struct dentry *) * sbmax(dentry->d_sb);
 
 	/* Don't reallocate when we already have enough space. */
 	/* It would be ideal if we could actually use the slab macros to
@@ -440,7 +444,6 @@ int new_dentry_private_data(struct dentry *dentry)
 	if (oldsize) {
 		int minsize = malloc_sizes[0].cs_size;
 
-		PASSERT(dtohd_ptr(dentry));
 		if (!newsize || ((oldsize < newsize) && (newsize > minsize))) {
 			KFREE(dtohd_ptr(dentry));
 			dtohd_ptr(dentry) = NULL;
@@ -448,24 +451,23 @@ int new_dentry_private_data(struct dentry *dentry)
 	}
 
 	if (!dtohd_ptr(dentry) && newsize) {
-		dtohd_ptr(dentry) = KMALLOC(newsize, GFP_UNIONFS);
+		dtohd_ptr(dentry) = KMALLOC(newsize, GFP_ATOMIC);
 		if (!dtohd_ptr(dentry))
 			goto out;
 	}
-	memset(dtohd_inline(dentry), 0,
-	       UNIONFS_INLINE_OBJECTS * sizeof(struct dentry *));
-	if (newsize) {
-		if (oldsize > newsize)
-			memset(dtohd_ptr(dentry), 0, oldsize);
-		else
-			memset(dtohd_ptr(dentry), 0, newsize);
-	}
 
+	if (oldsize > newsize)
+		memset(dtohd_ptr(dentry), 0, oldsize);
+	else
+		memset(dtohd_ptr(dentry), 0, newsize);
+
+	spin_unlock(&dentry->d_lock);
 	return 0;
 
       out:
 	free_dentry_private_data(dtopd_nocheck(dentry));
 	dtopd_lhs(dentry) = NULL;
+	spin_unlock(&dentry->d_lock);
 	return -ENOMEM;
 }
 
