@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,13 +16,22 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: main.c,v 1.133 2005/10/20 16:24:41 cwright Exp $
+ *  $Id: main.c,v 1.171 2006/06/25 23:57:43 jro Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 #include <linux/module.h>
+#include <linux/moduleparam.h>
 
+/* declarations added for "sparse" */
+extern void unionfs_kill_block_super(struct super_block *sb);
+
+/* declarations added for malloc_debugging */
+
+#ifdef FIST_MALLOC_DEBUG
+extern atomic_t unionfs_malloc_counter;
+extern atomic_t unionfs_mallocs_outstanding;
+#endif
 /* sb we pass is unionfs's super_block */
 int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 {
@@ -36,7 +46,7 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 
 	verify_locked(dentry);
 
-	fist_print_dentry("In unionfs_interpose", dentry);
+	print_dentry("In unionfs_interpose", dentry);
 
 	bstart = dbstart(dentry);
 	bend = dbend(dentry);
@@ -49,7 +59,7 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 			break;
 		}
 	}
-	ASSERT(!is_negative_dentry);
+	BUG_ON(is_negative_dentry);
 
 	/* We allocate our new inode below, by calling iget.
 	 * iget will call our read_inode which will initialize some
@@ -60,41 +70,42 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 	 * to fix it up. */
 	if (flag == INTERPOSE_REVAL) {
 		inode = dentry->d_inode;
-		PASSERT(inode);
 		itopd(inode)->b_start = -1;
 		itopd(inode)->b_end = -1;
 		atomic_set(&itopd(inode)->uii_generation,
 			   atomic_read(&stopd(sb)->usi_generation));
 
-		if (sbmax(sb) > UNIONFS_INLINE_OBJECTS) {
-			int size =
-			    (sbmax(sb) -
-			     UNIONFS_INLINE_OBJECTS) * sizeof(struct inode *);
-			itohi_ptr(inode) = KMALLOC(size, GFP_UNIONFS);
-			if (!itohi_ptr(inode)) {
-				err = -ENOMEM;
-				goto out;
-			}
-			memset(itohi_ptr(inode), 0, size);
+		itohi_ptr(inode) =
+		    KZALLOC(sbmax(sb) * sizeof(struct inode *), GFP_KERNEL);
+		if (!itohi_ptr(inode)) {
+			err = -ENOMEM;
+			goto out;
 		}
-		memset(itohi_inline(inode), 0,
-		       UNIONFS_INLINE_OBJECTS * sizeof(struct inode *));
+		mutex_lock(&inode->i_mutex);
 	} else {
 		ino_t ino;
 		/* get unique inode number for unionfs */
-		if (stopd(sb)->usi_persistent)
-			ino =
-			    get_uin(sb, bindex,
-				    dtohd_index(dentry, bindex)->d_inode->i_ino,
-				    O_CREAT);
-		else
+#ifdef UNIONFS_IMAP
+		if (stopd(sb)->usi_persistent) {
+			err = read_uin(sb, bindex,
+				       dtohd_index(dentry,
+						   bindex)->d_inode->i_ino,
+				       O_CREAT, &ino);
+			if (err)
+				goto out;
+		} else
+#endif
 			ino = iunique(sb, UNIONFS_ROOT_INO);
 
-		inode = iget(sb, ino);
+		inode = IGET(sb, ino);
 		if (!inode) {
 			err = -EACCES;	/* should be impossible??? */
 			goto out;
 		}
+
+		mutex_lock(&inode->i_mutex);
+		if (atomic_read(&inode->i_count) > 1)
+			goto skip;
 	}
 
 	for (bindex = bstart; bindex <= bend; bindex++) {
@@ -106,7 +117,7 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 		/* Initialize the hidden inode to the new hidden inode. */
 		if (!hidden_dentry->d_inode)
 			continue;
-		set_itohi_index(inode, bindex, igrab(hidden_dentry->d_inode));
+		set_itohi_index(inode, bindex, IGRAB(hidden_dentry->d_inode));
 	}
 
 	ibstart(inode) = dbstart(dentry);
@@ -114,7 +125,6 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 
 	/* Use attributes from the first branch. */
 	hidden_inode = itohi(inode);
-	PASSERT(hidden_inode);
 
 	/* Use different set of inode ops for symlinks & directories */
 	if (S_ISLNK(hidden_inode->i_mode))
@@ -131,18 +141,19 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 	    S_ISFIFO(hidden_inode->i_mode) || S_ISSOCK(hidden_inode->i_mode))
 		init_special_inode(inode, hidden_inode->i_mode,
 				   hidden_inode->i_rdev);
-
+#ifndef UNIONFS_MMAP
 	/* Fix our inode's address operations to that of the lower inode (Unionfs is FiST-Lite) */
 	if (inode->i_mapping->a_ops != hidden_inode->i_mapping->a_ops) {
-		fist_dprint(7, "fixing inode 0x%p a_ops (0x%p -> 0x%p)\n",
-			    inode, inode->i_mapping->a_ops,
-			    hidden_inode->i_mapping->a_ops);
+		dprint(PRINT_DEBUG, "fixing inode 0x%p a_ops (0x%p -> 0x%p)\n",
+		       inode, inode->i_mapping->a_ops,
+		       hidden_inode->i_mapping->a_ops);
 		inode->i_mapping->a_ops = hidden_inode->i_mapping->a_ops;
 	}
-
+#endif
 	/* all well, copy inode attributes */
 	fist_copy_attr_all(inode, hidden_inode);
 
+      skip:
 	/* only (our) lookup wants to do a d_add */
 	switch (flag) {
 	case INTERPOSE_DEFAULT:
@@ -156,13 +167,13 @@ int unionfs_interpose(struct dentry *dentry, struct super_block *sb, int flag)
 		/* Do nothing. */
 		break;
 	default:
-		FISTBUG("Invalid interpose flag passed!");
+		printk(KERN_ERR "Invalid interpose flag passed!");
+		BUG();
 	}
 
-	PASSERT(dtopd(dentry));
-
-	fist_print_dentry("Leaving unionfs_interpose", dentry);
-	fist_print_inode("Leaving unionfs_interpose", inode);
+	print_dentry("Leaving unionfs_interpose", dentry);
+	print_inode("Leaving unionfs_interpose", inode);
+	mutex_unlock(&inode->i_mutex);
 
       out:
 	print_exit_status(err);
@@ -177,11 +188,10 @@ void unionfs_reinterpose(struct dentry *dentry)
 
 	print_entry_location();
 	verify_locked(dentry);
-	fist_print_dentry("IN: unionfs_reinterpose: ", dentry);
+	print_dentry("IN: unionfs_reinterpose: ", dentry);
 
 	/* This is pre-allocated inode */
 	inode = dentry->d_inode;
-	PASSERT(inode);
 
 	bstart = dbstart(dentry);
 	bend = dbend(dentry);
@@ -194,23 +204,19 @@ void unionfs_reinterpose(struct dentry *dentry)
 			continue;
 		if (itohi_index(inode, bindex))
 			continue;
-		set_itohi_index(inode, bindex, igrab(hidden_dentry->d_inode));
+		set_itohi_index(inode, bindex, IGRAB(hidden_dentry->d_inode));
 	}
 	ibstart(inode) = dbstart(dentry);
 	ibend(inode) = dbend(dentry);
 
-	fist_print_dentry("OUT: unionfs_reinterpose: ", dentry);
-	fist_print_inode("OUT: unionfs_reinterpose: ", inode);
+	print_dentry("OUT: unionfs_reinterpose: ", dentry);
+	print_inode("OUT: unionfs_reinterpose: ", inode);
 
 	print_exit_location();
 }
 
 int check_branch(struct nameidata *nd)
 {
-	PASSERT(nd);
-	PASSERT(nd->dentry);
-	PASSERT(nd->dentry->d_sb);
-	PASSERT(nd->dentry->d_sb->s_type);
 	if (!strcmp(nd->dentry->d_sb->s_type->name, "unionfs"))
 		return -EINVAL;
 	if (!nd->dentry->d_inode)
@@ -220,16 +226,62 @@ int check_branch(struct nameidata *nd)
 	return 0;
 }
 
+/* checks if two hidden_dentries have overlapping branches */
+int is_branch_overlap(struct dentry *dent1, struct dentry *dent2)
+{
+	struct dentry *dent = NULL;
+
+	dent = dent1;
+	while ((dent != dent2) && (dent->d_parent != dent)) {
+		dent = dent->d_parent;
+	}
+	if (dent == dent2) {
+		return 1;
+	}
+
+	dent = dent2;
+	while ((dent != dent1) && (dent->d_parent != dent)) {
+		dent = dent->d_parent;
+	}
+	if (dent == dent1) {
+		return 1;
+	}
+
+	return 0;
+}
+static int parse_branch_mode(char *name)
+{
+	int perms;
+	int l = strlen(name);
+	if (!strcmp(name + l - 3, "=ro")) {
+		perms = MAY_READ;
+		name[l - 3] = '\0';
+	} else if (!strcmp(name + l - 6, "=nfsro")) {
+		perms = MAY_READ | MAY_NFSRO;
+		name[l - 6] = '\0';
+	} else if (!strcmp(name + l - 3, "=rw")) {
+		perms = MAY_READ | MAY_WRITE;
+		name[l - 3] = '\0';
+	} else {
+		perms = MAY_READ | MAY_WRITE;
+	}
+
+	return perms;
+}
+
 static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
 			     *hidden_root_info, char *options)
 {
 	struct nameidata nd;
 	char *name;
-	int pobjects = 0;
 	int err = 0;
 	int branches = 1;
 	int bindex = 0;
-	int i;
+	int i = 0;
+	int j = 0;
+
+	struct dentry *dent1 = NULL;
+	struct dentry *dent2 = NULL;
 
 	if (options[0] == '\0') {
 		printk(KERN_WARNING "unionfs: no branches specified\n");
@@ -244,54 +296,21 @@ static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
 			branches++;
 	}
 
-	/* allocate space for  underlying pointers to hidden dentry */
-	if (branches > UNIONFS_INLINE_OBJECTS)
-		pobjects = branches - UNIONFS_INLINE_OBJECTS;
-
-	if (pobjects) {
+	/* allocate space for underlying pointers to hidden dentry */
+	if (!(stopd(sb)->usi_data = alloc_new_data(branches))) {
 		err = -ENOMEM;
+		goto out;
+	}
 
-		hidden_root_info->udi_dentry_p =
-		    KMALLOC(sizeof(struct dentry *) * pobjects, GFP_UNIONFS);
-		if (!hidden_root_info->udi_dentry_p)
-			goto out;
-
-		memset(hidden_root_info->udi_dentry_p, 0,
-		       sizeof(struct dentry *) * pobjects);
-
-		stohs_ptr(sb) =
-		    KMALLOC(sizeof(struct super_block *) *
-			    pobjects, GFP_UNIONFS);
-		if (!stohs_ptr(sb))
-			goto out;
-
-		memset(stohs_ptr(sb), 0,
-		       pobjects * sizeof(struct super_block *));
-
-		stopd(sb)->usi_sbcount_p =
-		    KMALLOC(sizeof(atomic_t) * pobjects, GFP_UNIONFS);
-		if (!stopd(sb)->usi_sbcount_p)
-			goto out;
-
-		stopd(sb)->usi_branchperms_p =
-		    KMALLOC(sizeof(int) * pobjects, GFP_UNIONFS);
-		if (!stopd(sb)->usi_branchperms_p)
-			goto out;
-
-		stohiddenmnt_ptr(sb) =
-		    KMALLOC(sizeof(struct vfsmount *) * pobjects, GFP_UNIONFS);
-		if (!stohiddenmnt_ptr(sb))
-			goto out;
-
-		/* We're done with allocating memory. */
-		err = 0;
+	if (!(hidden_root_info->udi_dentry = alloc_new_dentries(branches))) {
+		err = -ENOMEM;
+		goto out;
 	}
 
 	/* now parsing the string b1:b2=rw:b3=ro:b4 */
 	branches = 0;
 	while ((name = strsep(&options, ":")) != NULL) {
 		int perms;
-		int l;
 
 		if (!*name)
 			continue;
@@ -299,20 +318,16 @@ static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
 		branches++;
 
 		/* strip off =rw or =ro if it is specified. */
-		l = strlen(name);
-		if (!strcmp(name + l - 3, "=ro")) {
-			perms = MAY_READ;
-			name[l - 3] = '\0';
-		} else if (!strcmp(name + l - 3, "=rw")) {
-			perms = MAY_READ | MAY_WRITE;
-			name[l - 3] = '\0';
-		} else {
-			perms = MAY_READ | MAY_WRITE;
+		perms = parse_branch_mode(name);
+		if (!bindex && !(perms & MAY_WRITE)) {
+			err = -EINVAL;
+			goto out;
 		}
 
-		fist_dprint(4, "unionfs: using directory: %s (%c%c)\n",
-			    name, perms & MAY_READ ? 'r' : '-',
-			    perms & MAY_WRITE ? 'w' : '-');
+		dprint(PRINT_DEBUG, "using directory: %s (%c%c%c)\n",
+		       name, perms & MAY_READ ? 'r' : '-',
+		       perms & MAY_WRITE ? 'w' : '-',
+		       perms & MAY_NFSRO ? 'n' : '-');
 
 		err = path_lookup(name, LOOKUP_FOLLOW, &nd);
 		RECORD_PATH_LOOKUP(&nd);
@@ -330,12 +345,7 @@ static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
 			goto out;
 		}
 
-		if (bindex < UNIONFS_INLINE_OBJECTS)
-			hidden_root_info->udi_dentry_i[bindex] = nd.dentry;
-		else
-			hidden_root_info->udi_dentry_p[bindex -
-						       UNIONFS_INLINE_OBJECTS]
-			    = nd.dentry;
+		hidden_root_info->udi_dentry[bindex] = nd.dentry;
 
 		set_stohiddenmnt_index(sb, bindex, nd.mnt);
 		set_branchperms(sb, bindex, perms);
@@ -346,14 +356,52 @@ static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
 		hidden_root_info->udi_bend = bindex;
 		bindex++;
 	}
+
 	if (branches == 0) {
 		printk(KERN_WARNING "unionfs: no branches specified\n");
 		err = -EINVAL;
 		goto out;
 	}
 
-	ASSERT(branches == (hidden_root_info->udi_bend + 1));
+	BUG_ON(branches != (hidden_root_info->udi_bend + 1));
+
+	/* ensure that no overlaps exist in the branches */
+	for (i = 0; i < branches; i++) {
+		for (j = i + 1; j < branches; j++) {
+			dent1 = hidden_root_info->udi_dentry[i];
+			dent2 = hidden_root_info->udi_dentry[j];
+
+			if (is_branch_overlap(dent1, dent2)) {
+				goto out_overlap;
+			}
+		}
+	}
+
+      out_overlap:
+
+	if (i != branches) {
+		printk(KERN_WARNING "unionfs: branches %d and %d overlap\n", i,
+		       j);
+		err = -EINVAL;
+		goto out;
+	}
+
       out:
+	if (err) {
+		for (i = 0; i < branches; i++) {
+			if (hidden_root_info->udi_dentry[i])
+				DPUT(hidden_root_info->udi_dentry[i]);
+		}
+
+		KFREE(hidden_root_info->udi_dentry);
+		KFREE(stopd(sb)->usi_data);
+
+		/* MUST clear the pointers to prevent potential double free if
+		 * the caller dies later on
+		 */
+		hidden_root_info->udi_dentry = NULL;
+		stopd(sb)->usi_data = NULL;
+	}
 	return err;
 }
 
@@ -365,25 +413,25 @@ static int parse_dirs_option(struct super_block *sb, struct unionfs_dentry_info
  *
  * Sets default debugging level to N, if any.
  */
-struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
-						  char *options)
+static struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
+							 char *options)
 {
 	struct unionfs_dentry_info *hidden_root_info;
 	char *optname;
 	int err = 0;
 	int bindex;
 	int dirsfound = 0;
+#ifdef UNIONFS_IMAP
 	int imapfound = 0;
-	int mounter_f = 0, copyupuid_f = 0, copyupgid_f = 0, copyupmode_f = 0;
+#endif
 	print_entry_location();
 
 	/* allocate private data area */
 	err = -ENOMEM;
 	hidden_root_info =
-	    KMALLOC(sizeof(struct unionfs_dentry_info), GFP_UNIONFS);
+	    KZALLOC(sizeof(struct unionfs_dentry_info), GFP_KERNEL);
 	if (!hidden_root_info)
 		goto out_error;
-	memset(hidden_root_info, 0, sizeof(struct unionfs_dentry_info));
 	hidden_root_info->udi_bstart = -1;
 	hidden_root_info->udi_bend = -1;
 	hidden_root_info->udi_bopaque = -1;
@@ -422,6 +470,7 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 				goto out_error;
 			continue;
 		}
+#ifdef UNIONFS_IMAP
 		if (!strcmp("imap", optname)) {
 			if (++imapfound > 1) {
 				printk(KERN_WARNING
@@ -434,11 +483,14 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 				goto out_error;
 			continue;
 		}
+#endif
 		if (!strcmp("delete", optname)) {
-			if (!strcmp("all", optarg)) {
+			if (!strcmp("whiteout", optarg)) {
 				/* default */
-			} else if (!strcmp("whiteout", optarg)) {
-				MOUNT_FLAG(sb) |= DELETE_WHITEOUT;
+#ifdef UNIONFS_DELETE_ALL
+			} else if (!strcmp("all", optarg)) {
+				MOUNT_FLAG(sb) |= DELETE_ALL;
+#endif
 			} else {
 				printk(KERN_WARNING
 				       "unionfs: invalid delete option '%s'\n",
@@ -448,24 +500,6 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 			}
 			continue;
 		}
-		if (!strcmp("copyup", optname)) {
-			if (!strcmp("preserve", optarg)) {
-				/* default */
-			} else if (!strcmp("currentuser", optarg)) {
-				MOUNT_FLAG(sb) |= COPYUP_CURRENT_USER;
-			} else if (!strcmp("mounter", optarg)) {
-				MOUNT_FLAG(sb) |= COPYUP_FS_MOUNTER;
-				mounter_f = 1;
-			} else {
-				printk(KERN_WARNING
-				       "unionfs: could not parse copyup option value '%s'\n",
-				       optarg);
-				err = -EINVAL;
-				goto out_error;
-			}
-			continue;
-		}
-
 		/* All of these options require an integer argument. */
 		intval = simple_strtoul(optarg, &endptr, 0);
 		if (*endptr) {
@@ -477,29 +511,7 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 		}
 
 		if (!strcmp("debug", optname)) {
-			fist_set_debug_value(intval);
-			continue;
-		}
-		if (!strcmp("copyupuid", optname)) {
-			stopd(sb)->copyupuid = intval;
-			copyupuid_f = 1;
-			continue;
-		}
-		if (!strcmp("copyupgid", optname)) {
-			stopd(sb)->copyupgid = intval;
-			copyupgid_f = 1;
-			continue;
-		}
-		if (!strcmp("copyupmode", optname)) {
-			if (intval & ~0777) {
-				err = -EINVAL;
-				printk(KERN_WARNING
-				       "unionfs: copyupmode invalid '%o' (note: you need to preface octal values with 0.\n",
-				       intval);
-				goto out_error;
-			}
-			stopd(sb)->copyupmode = intval;
-			copyupmode_f = 1;
+			set_debug_mask(intval);
 			continue;
 		}
 
@@ -513,47 +525,26 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 		err = -EINVAL;
 		goto out_error;
 	}
-	if (mounter_f && !(copyupuid_f && copyupgid_f && copyupmode_f)) {
-		printk(KERN_WARNING
-		       "unionfs: "
-		       "copyup=mounter all copyup options must be set\n");
-
-		err = -EINVAL;
-		goto out_error;
-	}
-	if ((copyupuid_f || copyupgid_f || copyupmode_f) && !mounter_f) {
-		printk(KERN_WARNING
-		       "unionfs: "
-		       "copyup!=mounter and a copyup option is set\n");
-		err = -EINVAL;
-		goto out_error;
-	}
 	goto out;
 
       out_error:
-	for (bindex = hidden_root_info->udi_bstart;
-	     bindex >= 0 && bindex <= hidden_root_info->udi_bend; bindex++) {
-		struct dentry *d;
-		if (bindex < UNIONFS_INLINE_OBJECTS)
-			d = hidden_root_info->udi_dentry_i[bindex];
-		else
-			d = hidden_root_info->udi_dentry_p[bindex -
-							   UNIONFS_INLINE_OBJECTS];
-		DPUT(d);
-		if (stohiddenmnt_index(sb, bindex))
-			mntput(stohiddenmnt_index(sb, bindex));
+	if (hidden_root_info && hidden_root_info->udi_dentry) {
+		for (bindex = hidden_root_info->udi_bstart;
+		     bindex >= 0 && bindex <= hidden_root_info->udi_bend;
+		     bindex++) {
+			struct dentry *d;
+			d = hidden_root_info->udi_dentry[bindex];
+			DPUT(d);
+			if (stohiddenmnt_index(sb, bindex))
+				mntput(stohiddenmnt_index(sb, bindex));
+		}
 	}
-	KFREE(hidden_root_info->udi_dentry_p);
+
+	KFREE(hidden_root_info->udi_dentry);
 	KFREE(hidden_root_info);
 
-	KFREE(stohiddenmnt_ptr(sb));
-	stohiddenmnt_ptr(sb) = NULL;
-	KFREE(stopd(sb)->usi_sbcount_p);
-	stopd(sb)->usi_sbcount_p = NULL;
-	KFREE(stopd(sb)->usi_branchperms_p);
-	stopd(sb)->usi_branchperms_p = NULL;
-	KFREE(stohs_ptr(sb));
-	stohs_ptr(sb) = NULL;
+	KFREE(stopd(sb)->usi_data);
+	stopd(sb)->usi_data = NULL;
 
 	hidden_root_info = ERR_PTR(err);
       out:
@@ -561,157 +552,25 @@ struct unionfs_dentry_info *unionfs_parse_options(struct super_block *sb,
 	return hidden_root_info;
 }
 
-#ifdef FIST_MALLOC_DEBUG
-/* for malloc debugging */
-atomic_t unionfs_malloc_counter;
-atomic_t unionfs_mallocs_outstanding;
-atomic_t unionfs_dget_counter;
-atomic_t unionfs_dgets_outstanding;
-
-void *unionfs_kmalloc(size_t len, int flag, int line, const char *file)
+static struct dentry *unionfs_d_alloc_root(struct super_block *sb)
 {
-	void *ptr = (void *)kmalloc(len, flag);
-	if (ptr) {
-		atomic_inc(&unionfs_malloc_counter);
-		atomic_inc(&unionfs_mallocs_outstanding);
-		printk("KM:%d:%d:%p:%d:%s\n",
-		       atomic_read(&unionfs_malloc_counter),
-		       atomic_read(&unionfs_mallocs_outstanding), ptr, line,
-		       file);
+	struct dentry *ret = NULL;
+
+	if (sb) {
+		static const struct qstr name = {.name = "/",.len = 1 };
+
+		ret = d_alloc(NULL, &name);
+		if (ret) {
+			ret->d_op = &unionfs_dops;
+			ret->d_sb = sb;
+			ret->d_parent = ret;
+		}
 	}
-	return ptr;
+	return ret;
 }
 
-void unionfs_kfree(void *ptr, int line, const char *file)
-{
-	atomic_inc(&unionfs_malloc_counter);
-	if (ptr) {
-		ASSERT(!IS_ERR(ptr));
-		atomic_dec(&unionfs_mallocs_outstanding);
-	}
-	printk("KF:%d:%d:%p:%d:%s\n", atomic_read(&unionfs_malloc_counter),
-	       atomic_read(&unionfs_mallocs_outstanding), ptr, line, file);
-	kfree(ptr);
-}
-
-void record_set(struct dentry *upper, int index, struct dentry *ptr,
-		struct dentry *old, int line, const char *file)
-{
-	atomic_inc(&unionfs_dget_counter);
-	printk("DD:%d:%d:%d:%p:%d:%s %p, %d\n",
-	       atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       old ? atomic_read(&old->d_count) : 0, old, line, file, upper,
-	       index);
-	atomic_inc(&unionfs_dget_counter);
-	printk("DS:%d:%d:%d:%p:%d:%s %p, %d\n",
-	       atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       ptr ? atomic_read(&ptr->d_count) : 0, ptr, line, file, upper,
-	       index);
-}
-
-void record_path_lookup(struct nameidata *nd, int line, const char *file)
-{
-	struct dentry *ptr = nd->dentry;
-	if (ptr) {
-		atomic_inc(&unionfs_dget_counter);
-		atomic_inc(&unionfs_dgets_outstanding);
-		printk("DL:%d:%d:%d:%p:%d:%s\n",
-		       atomic_read(&unionfs_dget_counter),
-		       atomic_read(&unionfs_dgets_outstanding),
-		       atomic_read(&ptr->d_count), ptr, line, file);
-	}
-}
-
-void record_path_release(struct nameidata *nd, int line, const char *file)
-{
-	struct dentry *ptr = nd->dentry;
-
-	atomic_inc(&unionfs_dget_counter);
-	if (ptr)
-		atomic_dec(&unionfs_dgets_outstanding);
-	printk("DP:%d:%d:%d:%p:%d:%s\n", atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       ptr ? atomic_read(&ptr->d_count) : 0, ptr, line, file);
-}
-
-struct file *unionfs_dentry_open(struct dentry *ptr, struct vfsmount *mnt,
-				 int flags, int line, const char *file)
-{
-	atomic_inc(&unionfs_dget_counter);
-	if (ptr)
-		atomic_dec(&unionfs_dgets_outstanding);
-	printk("DO:%d:%d:%d:%p:%d:%s\n", atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       ptr ? atomic_read(&ptr->d_count) : 0, ptr, line, file);
-	return dentry_open(ptr, mnt, flags);
-}
-
-struct dentry *unionfs_dget(struct dentry *ptr, int line, const char *file)
-{
-	ptr = dget(ptr);
-	if (ptr) {
-		atomic_inc(&unionfs_dget_counter);
-		atomic_inc(&unionfs_dgets_outstanding);
-		printk("DG:%d:%d:%d:%p:%d:%s\n",
-		       atomic_read(&unionfs_dget_counter),
-		       atomic_read(&unionfs_dgets_outstanding),
-		       atomic_read(&ptr->d_count), ptr, line, file);
-	}
-	return ptr;
-}
-
-struct dentry *unionfs_dget_parent(struct dentry *child, int line,
-				   const char *file)
-{
-	struct dentry *ptr;
-
-	PASSERT(child);
-	ptr = dget_parent(child);
-	PASSERT(ptr);
-	atomic_inc(&unionfs_dget_counter);
-	atomic_inc(&unionfs_dgets_outstanding);
-	printk("DG:%d:%d:%d:%p:%d:%s\n",
-	       atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       atomic_read(&ptr->d_count), ptr, line, file);
-
-	return ptr;
-}
-
-struct dentry *unionfs_lookup_one_len(const char *name, struct dentry *parent,
-				      int len, int line, const char *file)
-{
-	struct dentry *ptr = lookup_one_len(name, parent, len);
-	if (ptr && !IS_ERR(ptr)) {
-		atomic_inc(&unionfs_dget_counter);
-		atomic_inc(&unionfs_dgets_outstanding);
-		printk("DL:%d:%d:%d:%p:%d:%s\n",
-		       atomic_read(&unionfs_dget_counter),
-		       atomic_read(&unionfs_dgets_outstanding),
-		       atomic_read(&ptr->d_count), ptr, line, file);
-	}
-	return ptr;
-}
-
-void unionfs_dput(struct dentry *ptr, int line, const char *file)
-{
-	atomic_inc(&unionfs_dget_counter);
-	if (ptr) {
-		ASSERT(!IS_ERR(ptr));
-		atomic_dec(&unionfs_dgets_outstanding);
-	}
-	printk("DP:%d:%d:%d:%p:%d:%s\n", atomic_read(&unionfs_dget_counter),
-	       atomic_read(&unionfs_dgets_outstanding),
-	       ptr ? atomic_read(&ptr->d_count) : 0, ptr, line, file);
-	dput(ptr);
-}
-
-#endif				/* FIST_MALLOC_DEBUG */
-
-static int
-unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
+static int unionfs_read_super(struct super_block *sb, void *raw_data,
+			      int silent)
 {
 	int err = 0;
 
@@ -731,13 +590,12 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 	/*
 	 * Allocate superblock private data
 	 */
-	stopd_lhs(sb) = KMALLOC(sizeof(struct unionfs_sb_info), GFP_UNIONFS);
+	stopd_lhs(sb) = KZALLOC(sizeof(struct unionfs_sb_info), GFP_KERNEL);
 	if (!stopd(sb)) {
 		printk(KERN_WARNING "%s: out of memory\n", __FUNCTION__);
 		err = -ENOMEM;
 		goto out;
 	}
-	memset(stopd(sb), 0, sizeof(struct unionfs_sb_info));
 	stopd(sb)->b_end = -1;
 	atomic_set(&stopd(sb)->usi_generation, 1);
 	init_rwsem(&stopd(sb)->usi_rwsem);
@@ -758,16 +616,12 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 
 	/* set the hidden superblock field of upper superblock */
 	bstart = hidden_root_info->udi_bstart;
-	ASSERT(bstart == 0);
+	BUG_ON(bstart != 0);
 	sbend(sb) = bend = hidden_root_info->udi_bend;
 	for (bindex = bstart; bindex <= bend; bindex++) {
 		struct dentry *d;
 
-		if (bindex < UNIONFS_INLINE_OBJECTS)
-			d = hidden_root_info->udi_dentry_i[bindex];
-		else
-			d = hidden_root_info->udi_dentry_p[bindex -
-							   UNIONFS_INLINE_OBJECTS];
+		d = hidden_root_info->udi_dentry[bindex];
 
 		set_stohs_index(sb, bindex, d->d_sb);
 	}
@@ -780,25 +634,21 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 	sb->s_maxbytes = maxbytes;
 
 	sb->s_op = &unionfs_sops;
+#ifdef CONFIG_EXPORTFS
 	sb->s_export_op = &unionfs_export_ops;
+#endif
 
 	/*
 	 * we can't use d_alloc_root if we want to use
 	 * our own interpose function unchanged,
-	 * so we simply replicate *most* of the code in d_alloc_root here
+	 * so we simply call our own "fake" d_alloc_root
 	 */
-	sb->s_root = d_alloc(NULL, &(const struct qstr) {
-      hash: 0, name: "/", len:1});
-	if (IS_ERR(sb->s_root)) {
-		printk(KERN_WARNING "unionfs_read_super: d_alloc failed\n");
-		err = PTR_ERR(sb->s_root);
+	sb->s_root = unionfs_d_alloc_root(sb);
+	if (!sb->s_root) {
+		err = -ENOMEM;
 		goto out_dput;
 	}
 
-	sb->s_root->d_op = &unionfs_dops;
-	sb->s_root->d_sb = sb;
-	sb->s_root->d_parent = sb->s_root;
-
 	/* link the upper and lower dentries */
 	dtopd_lhs(sb->s_root) = NULL;
 	if ((err = new_dentry_private_data(sb->s_root)))
@@ -808,11 +658,7 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 	for (bindex = bstart; bindex <= bend; bindex++) {
 		struct dentry *d;
 
-		if (bindex < UNIONFS_INLINE_OBJECTS)
-			d = hidden_root_info->udi_dentry_i[bindex];
-		else
-			d = hidden_root_info->udi_dentry_p[bindex -
-							   UNIONFS_INLINE_OBJECTS];
+		d = hidden_root_info->udi_dentry[bindex];
 
 		set_dtohd_index(sb->s_root, bindex, d);
 	}
@@ -840,13 +686,7 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 		     bindex <= hidden_root_info->udi_bend; bindex++) {
 			struct dentry *d;
 
-			if (bindex < UNIONFS_INLINE_OBJECTS)
-				d = hidden_root_info->udi_dentry_i[bindex];
-			else if (!hidden_root_info->udi_dentry_p)
-				break;
-			else
-				d = hidden_root_info->udi_dentry_p[bindex -
-								   UNIONFS_INLINE_OBJECTS];
+			d = hidden_root_info->udi_dentry[bindex];
 
 			if (d)
 				DPUT(d);
@@ -854,20 +694,17 @@ unionfs_read_super(struct super_block *sb, void *raw_data, int silent)
 			if (stopd(sb) && stohiddenmnt_index(sb, bindex))
 				mntput(stohiddenmnt_index(sb, bindex));
 		}
-		KFREE(hidden_root_info->udi_dentry_p);
+		KFREE(hidden_root_info->udi_dentry);
 		KFREE(hidden_root_info);
 		hidden_root_info = NULL;
 	}
       out_free:
-	KFREE(stohiddenmnt_ptr(sb));
-	KFREE(stopd(sb)->usi_sbcount_p);
-	KFREE(stopd(sb)->usi_branchperms_p);
-	KFREE(stohs_ptr(sb));
+	KFREE(stopd(sb)->usi_data);
 	KFREE(stopd(sb));
 	stopd_lhs(sb) = NULL;
       out:
 	if (hidden_root_info && !IS_ERR(hidden_root_info)) {
-		KFREE(hidden_root_info->udi_dentry_p);
+		KFREE(hidden_root_info->udi_dentry);
 		KFREE(hidden_root_info);
 	}
 	print_exit_status(err);
@@ -894,11 +731,17 @@ static struct file_system_type unionfs_fs_type = {
 	.fs_flags = FS_REVAL_DOT,
 };
 
+static int init_debug = 0;
+module_param_named(debug, init_debug, int, S_IRUGO);
+MODULE_PARM_DESC(debug, "Initial Unionfs debug value.");
+
 static int __init init_unionfs_fs(void)
 {
 	int err;
 	printk("Registering unionfs " UNIONFS_VERSION "\n");
 
+	set_debug_mask(init_debug);
+
 #ifdef FIST_MALLOC_DEBUG
 	atomic_set(&unionfs_malloc_counter, 0);
 	atomic_set(&unionfs_mallocs_outstanding, 0);
@@ -928,8 +771,10 @@ static void __exit exit_unionfs_fs(void)
 	printk("Completed unionfs module unload.\n");
 }
 
-MODULE_AUTHOR("Erez Zadok <ezk@cs.columbia.edu>");
-MODULE_DESCRIPTION("FiST-generated unionfs filesystem");
+MODULE_AUTHOR
+    ("Filesystems and Storage Lab, Stony Brook University (http://www.fsl.cs.sunysb.edu/)");
+MODULE_DESCRIPTION("Unionfs " UNIONFS_VERSION
+		   " (http://unionfs.filesystems.org/)");
 MODULE_LICENSE("GPL");
 
 module_init(init_unionfs_fs);
