@@ -1,13 +1,14 @@
 /*
- * Copyright (c) 2003-2005 Erez Zadok
- * Copyright (c) 2003-2005 Charles P. Wright
+ * Copyright (c) 2003-2006 Erez Zadok
+ * Copyright (c) 2003-2006 Charles P. Wright
+ * Copyright (c) 2005-2006 Josef Sipek
  * Copyright (c) 2005      Arun M. Krishnakumar
- * Copyright (c) 2005      David P. Quigley
+ * Copyright (c) 2005-2006 David P. Quigley
  * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
- * Copyright (c) 2003-2003 Puja Gupta
- * Copyright (c) 2003-2003 Harikesavan Krishnan
- * Copyright (c) 2003-2005 Stony Brook University
- * Copyright (c) 2003-2005 The Research Foundation of State University of New York
+ * Copyright (c) 2003      Puja Gupta
+ * Copyright (c) 2003      Harikesavan Krishnan
+ * Copyright (c) 2003-2006 Stony Brook University
+ * Copyright (c) 2003-2006 The Research Foundation of State University of New York
  *
  * For specific licensing information, see the COPYING file distributed with
  * this package.
@@ -15,10 +16,9 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: rdstate.c,v 1.25 2005/09/18 05:02:56 jsipek Exp $
+ *  $Id: rdstate.c,v 1.33 2006/06/01 03:11:03 jsipek Exp $
  */
 
-#include "fist.h"
 #include "unionfs.h"
 
 /* This file contains the routines for maintaining readdir state. */
@@ -30,19 +30,19 @@
  * by the dentry structure), then we use an inline name to preserve kmalloc
  * space. */
 static kmem_cache_t *unionfs_filldir_cachep;
-int init_filldir_cache()
+int init_filldir_cache(void)
 {
 	unionfs_filldir_cachep =
 	    kmem_cache_create("unionfs_filldir", sizeof(struct filldir_node), 0,
 			      SLAB_RECLAIM_ACCOUNT, NULL, NULL);
 
-	if (!unionfs_filldir_cachep) {
+	if (!unionfs_filldir_cachep)
 		return -ENOMEM;
-	}
+
 	return 0;
 }
 
-void destroy_filldir_cache()
+void destroy_filldir_cache(void)
 {
 	if (!unionfs_filldir_cachep)
 		return;
@@ -65,7 +65,7 @@ void destroy_filldir_cache()
 #define DENTPERONEPAGE 12
 #define DENTPERPAGE 118
 #define MINHASHSIZE 1
-int guesstimate_hash_size(struct inode *inode)
+static int guesstimate_hash_size(struct inode *inode)
 {
 	struct inode *hidden_inode;
 	int bindex;
@@ -91,11 +91,8 @@ int guesstimate_hash_size(struct inode *inode)
 
 int init_rdstate(struct file *file)
 {
-	ASSERT(sizeof(loff_t) == (sizeof(unsigned int) + sizeof(unsigned int)));
-	PASSERT(file);
-	ASSERT(ftopd(file)->rdstate == NULL);
-	PASSERT(file->f_dentry);
-	PASSERT(file->f_dentry->d_inode);
+	BUG_ON(sizeof(loff_t) != (sizeof(unsigned int) + sizeof(unsigned int)));
+	BUG_ON(ftopd(file)->rdstate != NULL);
 
 	ftopd(file)->rdstate =
 	    alloc_rdstate(file->f_dentry->d_inode, fbstart(file));
@@ -152,7 +149,7 @@ struct unionfs_dir_state *alloc_rdstate(struct inode *inode, int bindex)
 	    (mallocsize -
 	     sizeof(struct unionfs_dir_state)) / sizeof(struct list_head);
 
-	rdstate = KMALLOC(mallocsize, GFP_UNIONFS);
+	rdstate = KMALLOC(mallocsize, GFP_KERNEL);
 	if (!rdstate)
 		return NULL;
 
@@ -176,9 +173,8 @@ struct unionfs_dir_state *alloc_rdstate(struct inode *inode, int bindex)
 	return rdstate;
 }
 
-void free_filldir_node(struct filldir_node *node)
+static void free_filldir_node(struct filldir_node *node)
 {
-	PASSERT(node);
 	if (node->namelen >= DNAME_INLINE_LEN_MIN)
 		KFREE(node->name);
 	kmem_cache_free(unionfs_filldir_cachep, node);
@@ -189,8 +185,6 @@ void free_rdstate(struct unionfs_dir_state *state)
 	struct filldir_node *tmp;
 	int i;
 
-	PASSERT(state);
-
 	for (i = 0; i < state->uds_size; i++) {
 		struct list_head *head = &(state->uds_list[i]);
 		struct list_head *pos, *n;
@@ -220,9 +214,7 @@ struct filldir_node *find_filldir_node(struct unionfs_dir_state *rdstate,
 	/* print_entry("name = %*s", namelen, name); */
 	print_entry_location();
 
-	PASSERT(rdstate);
-	PASSERT(name);
-	ASSERT(namelen > 0);
+	BUG_ON(namelen <= 0);
 
 	hash = full_name_hash(name, namelen);
 	index = hash % rdstate->uds_size;
@@ -238,7 +230,7 @@ struct filldir_node *find_filldir_node(struct unionfs_dir_state *rdstate,
 			/* if the duplicate is in this branch, then the file system is corrupted. */
 			if (cursor->bindex == rdstate->uds_bindex) {
 				//buf->error = err = -EIO;
-				fist_dprint(8,
+				dprint(PRINT_DEBUG,
 					    "Possible I/O error unionfs_filldir: a file is duplicated in the same branch %d: %s\n",
 					    rdstate->uds_bindex, cursor->name);
 			}
@@ -281,9 +273,7 @@ int add_filldir_node(struct unionfs_dir_state *rdstate, const char *name,
 	/* print_entry("name = %*s", namelen, name); */
 	print_entry_location();
 
-	PASSERT(rdstate);
-	PASSERT(name);
-	ASSERT(namelen > 0);
+	BUG_ON(namelen <= 0);
 
 	hash = full_name_hash(name, namelen);
 	index = hash % rdstate->uds_size;
@@ -304,7 +294,7 @@ int add_filldir_node(struct unionfs_dir_state *rdstate, const char *name,
 	if (namelen < DNAME_INLINE_LEN_MIN) {
 		new->name = new->iname;
 	} else {
-		new->name = (char *)KMALLOC(namelen + 1, GFP_UNIONFS);
+		new->name = (char *)KMALLOC(namelen + 1, GFP_KERNEL);
 		if (!new->name) {
 			kmem_cache_free(unionfs_filldir_cachep, new);
 			new = NULL;
