@@ -38,7 +38,10 @@
 # TEST:  | S| S| S| rI       |
 # TEST:  | D| D| D|          |
 # TEST:  +--------+----------+
-
+if [ -z "$DELETE_ALL" ] ; then
+	echo "$0: delete=all testing is not enabled."
+	exit 0
+fi
 source scaffold
 
 function files {
@@ -226,7 +229,7 @@ function ro {
 function rwro {
 	STATE=$1
 	( files ; beforefiles) | create_hierarchy
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=$STATE $LOWER_DIR/b2=$STATE
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=$STATE $LOWER_DIR/b2=$STATE
 	for X in A B C D E F G H
 	do
 		mv -f "$MOUNTPOINT/r$X.S" "$MOUNTPOINT/r$X.D"
@@ -241,7 +244,7 @@ function rwro {
 
 function immutable {
 	( files ; beforefiles_i) | create_hierarchy
-	mount_union "" $LOWER_DIR/b?
+	mount_union "delete=all" $LOWER_DIR/b?
 	X=I
 	shouldfail mv -f "$MOUNTPOINT/r$X.S" "$MOUNTPOINT/r$X.D"
 	checktype "$MOUNTPOINT/r$X.S" 'f'
@@ -252,7 +255,7 @@ function immutable {
 
 function BUG388 {
 	( files ; beforefiles_388) | create_hierarchy
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
 	mv $MOUNTPOINT/d $MOUNTPOINT/d.new
 	mv $MOUNTPOINT/d.new $MOUNTPOINT/d
 	rm -rf $MOUNTPOINT/d
@@ -262,7 +265,7 @@ function BUG388 {
 
 function BUG425 {
 	( files ; beforefiles_425) | create_hierarchy
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
 	rmdir $MOUNTPOINT/e_dst
 	mv $MOUNTPOINT/f_src $MOUNTPOINT/e_dst
 	unmount_union
