@@ -13,7 +13,10 @@
 # TEST: Branches: b0,b1,b2
 # TEST: unlink(F)
 # TEST:  a file in b0,b1,b2 (b1 is immutable)
-
+if [ -z "$DELETE_ALL" ] ; then
+	echo "$0: delete=all testing is not enabled."	
+	exit 0
+fi
 source scaffold
 
 function files {
@@ -56,6 +59,13 @@ f $LOWER_DIR/b1/a
 FILES
 }
 
+function beforefiles_whcheck {
+cat <<FILES
+f $LOWER_DIR/b0/a
+f $LOWER_DIR/b1/.wh.a
+FILES
+}
+
 
 function afterfiles_ro {
 cat <<FILES
@@ -104,10 +114,16 @@ f $LOWER_DIR/b1/a
 FILES
 }
 
+function afterfiles_whcheck {
+cat <<FILES
+f $LOWER_DIR/b1/.wh.a
+FILES
+}
+
 function ro2 {
 	( files ; beforefiles) | create_hierarchy
 
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
 
 	/bin/unlink $MOUNTPOINT/d1/d2/e
 	checktype $MOUNTPOINT/d1/d2/e '-'
@@ -120,7 +136,7 @@ function ro2 {
 function ro {
 	( files ; beforefiles) | create_hierarchy
 
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
 
 	/bin/unlink $MOUNTPOINT/a
 	/bin/unlink $MOUNTPOINT/b
@@ -142,7 +158,7 @@ function ro {
 function rw {
 	( files ; beforefiles) | create_hierarchy
 
-	mount_union "" $LOWER_DIR/b[0-1]
+	mount_union "delete=all" $LOWER_DIR/b[0-1]
 
 	/bin/unlink $MOUNTPOINT/a
 	/bin/unlink $MOUNTPOINT/b
@@ -163,7 +179,7 @@ function rw {
 function immutable {
 	# The immutable test
 	( files ; beforefiles_i) | create_hierarchy
-	mount_union "" $LOWER_DIR/b[0-2]
+	mount_union "delete=all" $LOWER_DIR/b[0-2]
 	shouldfail /bin/unlink $MOUNTPOINT/f
 	checktype $MOUNTPOINT/f 'f'
 	unmount_union
@@ -172,7 +188,7 @@ function immutable {
 
 function BUG319 {
 	( files ; beforefiles_319) | create_hierarchy
-	mount_union "" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1=ro
 	checktype $MOUNTPOINT/a 'f'
 	rm -f $MOUNTPOINT/a
 	checktype $MOUNTPOINT/a '-'
@@ -184,8 +200,18 @@ function BUG319 {
 	( files ; afterfiles_319 )  | check_hierarchy $LOWER_DIR
 }
 
+function whcheck {
+	( files ; beforefiles_whcheck) | create_hierarchy
+	mount_union "delete=all" $LOWER_DIR/b0 $LOWER_DIR/b1
+	checktype $MOUNTPOINT/a 'f'
+	rm -f $MOUNTPOINT/a
+	checktype $MOUNTPOINT/a '-'
+	unmount_union
+	( files ; afterfiles_whcheck )  | check_hierarchy $LOWER_DIR
+}
+
 if [ -z "$FXNS" ] ; then
-	FXNS="rw ro BUG319"
+	FXNS="rw ro BUG319 whcheck"
 	if havechattr $LOWER_DIR ; then
 		FXNS="$FXNS immutable"
 	fi
